main :- 
   current_cycle(0), update_cycle(0),
   set_publisher_output_file,
   print("
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   Publisher cycle number 0  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



"), 
   my_published_protocol(WS_BPELProtocol),
   generate_publisher_code(WS_BPELProtocol, PrologProtocol, ProtocolId),
   assert_protocol_clauses(PrologProtocol),
   save_protocol_id(ProtocolId).

    
main :-
    current_cycle(X), X > 0, max_cycles(Max), X < Max, update_cycle(X),
    print("
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   Publisher cycle number "),
print(X),
print("  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



"),
    protocol_id(ProtocolId),
    print(
"
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   STAGE 3: execution of the protocol  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



"),  
print("******************** THE PROTOCOL STARTS ********************"),
nl, nl,    
exec(s(ProtocolId, 0)), 
nl, nl, 
print("*********** THE PROTOCOL TERMINATED SUCCESSFULLY ************"),
nl, nl. 
   
main :- true.     
     
set_publisher_output_file :- tell("last_publisher_output.txt").   


update_cycle(X) :-  X1 is X + 1,  retract(current_cycle(X):-true),  assert(current_cycle(X1):-true).
 
max_cycles(2):- true.
current_cycle(0) :- true. 
     
save_protocol_id(ProtocolId) :- assert(protocol_id(ProtocolId) :- true).
   
   %======================================================================================================================%
% GENERATION OF CODE

generate_publisher_code(
   process(
      %%% Definition of the mandatory parameters %%%
      parameters(
         ontology(O), 
         content_language(CL)),
            
      %%% Definition of the protocol name %%%
      protocol_name(ProtName),
      
      %%% Definition of the agent publisher %%%
      agent_publisher(short_name(PSN), long_name(PLN)),
      
      %%% Definition of the agent reader %%%
      agent_reader(short_name(RSN), long_name(RLN)),
     
      %%% Definition of the body contained in the WS_BPEL file %%%
      main_fragment(Body)
      ), 
      
      Program, 
      
      ProtName
) :-
print(
"
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%   STAGE 2: translation from the intermediate XML code to Prolog   %%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   Trace explaining the performed translation   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


"), 
 
generate_body(s(ProtName, 0), s(ProtName, final), RSN, Body, NestedProgram),
flatten(NestedProgram, ProgramWithoutEnd),
append(ProgramWithoutEnd, [exec(s(ProtName, final)) :- true],  Program),
print(
"


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% Generated Prolog program; indentation of clauses is driven by nesting in the original protocol %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

"),  
print_program(Program).

print_program([]) :- true.

print_program([Head|Tail]):-
extract_state_from_clause(Head, State),!,
pretty_print_clause(State, Head),
nl,nl, 
print_program(Tail).


print_program([Head|Tail]):-
print_program(Head),
nl,nl, 
print_program(Tail).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%% Definition of generate_body with final state to reach %%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% generate_body(
%%%            CurrentState,
%%%            StateToReach,
%%%            AtomicBPELActivity,
%%%            GeneratedPrologClause)

generate_body(
   s(Something, Index),
   SFinal, 
   Publisher,
   send(msg(Performative, Content)), 
   [exec(s(Something, Index)) :- (traceable_receive(Performative, Content, Publisher), exec(SFinal))]) 
   :-  
   debug(send,  
      s(Something, Index), SFinal, 
      send(msg(Performative, Content)), 
     [exec(s(Something, Index)) :- (traceable_receive(Performative, Content, Publisher), exec(SFinal))]).

generate_body(
   s(Something, Index), 
   SFinal,
   Publisher,
   receive(msg(Performative, Content)), 
   [exec(s(Something, Index)) :- 
            (traceable_send(Performative, Content, Publisher),              
             exec(SFinal))]) 
   :-    
   debug(receive, 
      s(Something, Index),SFinal,
      receive(msg(Performative, Content)), 
      [exec(s(Something, Index)) :- (traceable_send(Performative, Content, Publisher), 
             exec(SFinal))]).
   
   
generate_body(
   s(Something, Index), 
   SFinal,
   Publisher,
   while(Guard, While_activities),
   [ exec(s(Something, Index)) :- (check_while_guard_passed(s(Something, Index), Guard), !, exec(s(s(Something, Index), 0))) | 
     [ exec(s(Something, Index)) :- exec(SFinal) |
       [ exec(s(Something, Index1)) :- exec(s(Something, Index)) | While_code ]]] 
) :- 
   Index1 is Index + 1,
   debug(while, 
      s(Something, Index), SFinal, 
      while(Guard, While_activities),
      [ exec(s(Something, Index)) :- (check_while_guard_passed(s(Something, Index), Guard), !, exec(s(s(Something, Index), 0))) | 
     [ exec(s(Something, Index)) :- exec(SFinal) |
       [ exec(s(Something, Index1)) :- exec(s(Something, Index))]]]),
   generate_body(s(s(Something, Index), 0), s(Something, Index1), Publisher, While_activities, While_code).


generate_body(
   s(Something, Index), 
   SFinal, 
   Publisher,
   switch(Case_activities), 
   Case_code
) :- 
   debug(switch, 
   s(Something, Index), SFinal,
   switch(Case_activities), 
   []),
   generate_case_body(s(s(Something, Index),0), SFinal, Publisher, Case_activities, Case_code).
   
   
   
generate_body(
   s(Something, Index),
   SFinal,
   Publisher,
   if_then(Guard, Then_activities), 
   [ exec(s(Something, Index)) :- (check_if_guard_passed(s(Something, Index), Guard), !,  exec(s(s(Something, Index), 0))) | 
      [ exec(s(Something, Index)) :- exec(SFinal) | Then_code]]
) :- 
   debug(if_then,
    s(Something, Index),SFinal,
    if_then(Guard, Then_activities), 
    [ exec(s(Something, Index)) :- (check_if_guard_passed(s(Something, Index), Guard), !,  exec(s(s(Something, Index), 0))) | 
      [ exec(s(Something, Index)) :- exec(SFinal)]]),
    generate_body(s(s(Something, Index),0), SFinal, Publisher, Then_activities, Then_code).



generate_body(
   s(Something, Index),
   SFinal,
   Publisher,
   seq(Seq_activities), 
   [exec(s(Something, Index)) :- exec(s(s(Something, Index), 0)) | Seq_code]   
   ) :-  
   debug(sequence, s(Something, Index),SFinal,
   seq(Seq_activities), 
   [exec(s(Something, Index)) :- exec(s(s(Something, Index), 0))]),     
   generate_seq_body(
   s(s(Something, Index),0),
   SFinal, 
   Publisher,
   Seq_activities, 
   Seq_code).


generate_seq_body(
   s(Something, Index),
   SFinal, Publisher,
   [H|[]], 
   [H_code|[]]   
   ) :- !,
   generate_body(s(Something, Index), SFinal, Publisher, H, H_code).

generate_seq_body(
   s(Something, Index),
   SFinal, 
   Publisher,
   [H|T], 
   [H_code|Tail_code]   
   ) :- 
   Index1 is Index + 1, 
   generate_body(s(Something, Index), s(Something, Index1), Publisher, H, H_code),
   generate_seq_body(
   s(Something, Index1),
   SFinal, 
   Publisher,
   T, 
   Tail_code   
   ).
   
     
   

%%%%%%%%%%%%%%%%%%% Definition of generate_case_body %%%%%%%%%%%%%%%%%%%%%%%%%%%%%

generate_case_body(_, _, Publisher,  [], []) :- true.

generate_case_body(s(Something, CaseIndex), S1,  Publisher, [case(Guard, Activity)|Other_case_activities], 
[ exec(Something) :- (check_case_guard_passed(s(Something, CaseIndex), Guard), !,  exec(s(Something, CaseIndex))) | 
  Tail_case_code]
) :- 
   generate_body(s(Something, CaseIndex), S1, Publisher, Activity, Case_code),
   CaseIndex1 is CaseIndex + 1, 
   generate_case_body(s(Something, CaseIndex1), S1, Publisher, Other_case_activities, Other_case_code),
   append(Case_code, Other_case_code, Tail_case_code).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


debug(StatementName, State, FinalState, H, []):- !,
 pretty_print_state_nesting(State),print("Translating: "), print(StatementName), nl, 
 pretty_print_state_nesting(State),print("Current state: "), print(State), nl, 
 pretty_print_state_nesting(State),print("State to reach: "), print(FinalState), nl, 
 pretty_print_state_nesting(State),print("Statement under translation: "), print(H),nl, 
 pretty_print_state_nesting(State),print("Statement traslated: "), print("Indented statements below"), nl, nl. 

debug(StatementName, State, FinalState, H, HCode):-

 pretty_print_state_nesting(State),print("Translating: "), print(StatementName), nl, 
 pretty_print_state_nesting(State),print("Current state: "), print(State), nl, 
 pretty_print_state_nesting(State),print("State to reach: "), print(FinalState), nl, 
 pretty_print_state_nesting(State),print("Statement under translation: "), print(H),nl, 
 pretty_print_state_nesting(State),print("Statement traslated: "), print(HCode), nl, nl. 


pretty_print_state_nesting(s(Something, _)) :- !, print("  "), pretty_print_state_nesting(Something).
pretty_print_state_nesting(_) :- true.

extract_state_from_clause(exec(State) :- Body, State) :- !.
extract_state_from_clause(exec(State), State) :- true.



pretty_print_clause(State, exec(State) :- Body):-
pretty_print_state_nesting(State), print("exec(")
                                 , print(State),
                                   print(") :-"),nl,
pretty_print_body(State, Body).


pretty_print_body(State, (FirstAtom, Rest)) :-
!,
pretty_print_state_nesting(State),print("  "), print(FirstAtom),print(","), nl, 
pretty_print_body(State, Rest).

pretty_print_body(State, LastAtom) :-
pretty_print_state_nesting(State),print("  "), print(LastAtom),print("."), nl. 


flatten([], []):-!.

flatten([H|T], F) :- 
   !,
   flatten(H, FH),
   flatten(T, FT),
   append(FH, FT, F).
   
flatten(Elem, [Elem]) :- true.   
%======================================================================================================================%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Message reception management %%%%%%%%%%%%%%%%%%%%%%%

local_mailbox([]) :- true.

traceable_send(A, B, C) :- 
   send(A, B, C), !,
   print("Sent message: "), 
   print(A), print("  "), print(B), print("  "), print(C), print("  "), nl, nl.


traceable_receive(A, B, C) :- 
   receive(VA, VB, VC), 
   check_right_message(A, B, C, VA, VB, VC).
   
 
check_right_message(A, B, C, VA, VB, VC) :- 
   ((var(VA) ; var(VB)) ; var(VC)), !, 
   traceable_receive(A, B, C).  
 
check_right_message(A, B, C, VA, VB, VC) :- 
   (((A \= VA) ; (B \= VB)) ; C \= VC), !, 
   traceable_receive(A, B, C).  

check_right_message(A, B, C, VA, VB, VC) :- 
   print("Received message: "), 
   print(VA), print("  "), print(VB), print("  "), print(VC), print("  "), nl, nl.
   
   
   
check_while_guard_passed(s(Something, Index), Guard) :- 
   is_true(Guard), !, 
   print("Checked while guard: "), 
   print(s(Something, Index)), print("  "), print(Guard), print("  "), nl, nl.

check_if_guard_passed(s(Something, Index), Guard) :- 
   is_true(Guard), !, 
   print("Checked if guard: "), 
   print(s(Something, Index)), print("  "), print(Guard), print("  "), nl, nl.

check_case_guard_passed(s(Something, Index), Guard) :- 
   is_true(Guard), !, 
   print("Checked case guard: "), 
   print(s(Something, Index)), print("  "), print(Guard), print("  "), nl, nl.

is_true(no_guard) :- true.


member_and_remaining(Elem, List, RemainingList) :-
   member(Elem, List), 
   delete(Elem,List,RemainingList).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  

assert_protocol_clauses([H|T]) :-
   assert(H), 
   assert_protocol_clauses(T).
   
assert_protocol_clauses([]) :- true.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

my_published_protocol(          

%%%%%%%%% TBD
%%%%%%%%% CUT AND PASTE THE TERM THAT YOU FIND IN THE .FOL FILE GENERATED BY THE AUML2WSBPEL TRANSLATOR HERE! %%%%%%%%

%%%%%%%%% AFTER HAVING INSERTED THE TERM HERE, MAKE SURE YOU INSERT THE CORRECT JADE ADRRESSES OF BOTH THE 
%%%%%%%%% READER AND THE PUBLISHER INSIDE THE TERM. 
%%%%%%%%% YOU MUST EDIT THE SUBTERMS agent_publisher(short_name('publisher@address'),long_name('fs/fruitSeller:fruitSellerClass')),  
%%%%%%%%% AND agent_reader(short_name('reader@address'),long_name('fb/fruitBuyer:fruitBuyerClass'))
%%%%%%%%% AND SUBSTITUTE ONLY THE "address" STRING WITH THE STRING "name-of-your-computer:1099/JADE", WHERE
%%%%%%%%% OBVIOUSLY name-of-your-computer MUST BE ISTANTIATED WITH THE ACTUAL NAME OF YOUR COMPUTER

). 