main :- 
   current_cycle(0), update_cycle(0),
   set_reader_output_file,
   print("
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   Reader cycle number 0  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



"),
   look_for_a_new_useful_protocol(WS_BPELProtocol),
   syntax_correctness(WS_BPELProtocol, yes),
   generate_reader_code(WS_BPELProtocol, PrologProtocol, ProtocolId),
   assert_protocol_clauses(PrologProtocol),

%%%%%%%%%%%%%% TBD: Take care of the following goal: you should manually edit the condition to check, that is
%%%%%%%%%%%%%% currently empty (thus, it always succeeds)

   check_conditions(ProtocolId, [], Output),   
   save_output_for_the_condition_check(ProtocolId, Output).
  
   
main :- 
    current_cycle(X), X > 0, max_cycles(Max), X < Max, update_cycle(X),
    print("
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   Reader cycle number "),
print(X),
print("  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



"),
    decide_what_to_do(ProtocolId, Output).   
   
main :- true.    

update_cycle(X) :-  X1 is X + 1,  retract(current_cycle(X):-true),  assert(current_cycle(X1):-true).
 
max_cycles(2):- true.
current_cycle(0) :- true. 


save_output_for_the_condition_check(ProtocolId, Output) :- 
   assert(output_for_the_condition_check(ProtocolId, Output):-true).
   
set_reader_output_file :- tell("last_reader_output.txt").   

   
%%%%%%%%%%%%%%%%% decide what to do %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   
decide_what_to_do(ProtocolId, yes) :-
   output_for_the_condition_check(ProtocolId, yes),
   start_protocol(ProtocolId). 
   
decide_what_to_do(ProtocolId, no) :-
   output_for_the_condition_check(ProtocolId, no),
   retract_protocol_clauses(ProtocolId). 
   
start_protocol(ProtocolId):-
print(
"



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   STAGE 4: execution of the protocol   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



"),
print("******************** THE PROTOCOL STARTS ********************"),
nl, nl,    
exec(s(ProtocolId, 0)), 
nl, nl, 
print("*********** THE PROTOCOL TERMINATED SUCCESSFULLY ************"),
nl, nl.


retract_protocol_clauses(ProtocolId):-
print("**************  THE PROTOCOL IS NOT STARTED ****************"), nl, nl.   

%%%%%%%%%%%%%%%%% check conditions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   
check_conditions(ProtocolId, List, Res1) :-
print(
"



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   STAGE 3: verification of the conditions   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



"),    
check_conds(ProtocolId, List, Res1).
   
check_conds(ProtocolId, [HeadCond|TailCond], Res1):-
   check_one_cond(ProtocolId, HeadCond, Res), 
   go_on_checking(ProtocolId, TailCond, Res, Res1).  
   
go_on_checking(ProtocolId, TailCond, no, no) :- true.

go_on_checking(ProtocolId, TailCond, yes, Res1):-
   check_conds(ProtocolId, TailCond, Res1).      
   
check_conds(ProtocolId, [], yes) :- true.          
   
   
check_one_cond(ProtocolId, forall(C), no):-
   print(
"
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   Checking that ALL the possible paths meet  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  "),
write(C),
print(" 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


"),
   forall_solve(exec(s(ProtocolId, 0)), C, BadTrace), !, 
   nl, nl, 
   print("************** THE CONDITION IS NOT VERIFIED ***************"), 
   nl, nl. 
   
check_one_cond(ProtocolId, forall(C), yes) :- 
nl, nl, 
print("The above trace has been obtained by printing the actions that I simulated to
execute, and forcing a backtracking as soon as an action leading to a situation where
the condition was verified, was performed. By reasoning in this way, I was not able to reach 
the final state. This means that in any possible path, the condition was verified BEFORE the 
final state was reached. Namely, the condition was verified by ALL the possible paths. 

I basically performed a depth-first search with control of cycles, which is not as 
memory-consuming as a breadth-first, and is also more efficient.


*************** THE CONDITION IS VERIFIED *****************"), 
nl, nl.   
   

check_one_cond(ProtocolId, exists(C), yes):-
   print(
"
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%   Checking that AT LEAST ONE possible path meets  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%  "),
write(C),
print("  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


"),
   exists_solve(exec(s(ProtocolId, 0)), C, GoodTrace), !, 
   nl, nl, 
   print("The above trace has been obtained by printing the actions that I simulated to
execute, and forcing a backtracking as soon as the final state was reached before that
the condition was verified. 

I basically performed a depth-first search with control of cycles, which is not as 
memory-consuming as a breadth-first, and is also more efficient.


   *************** THE CONDITION IS VERIFIED *****************"), 
   nl, nl.
   
check_one_cond(ProtocolId, exists(C), no) :- 
   nl, nl, 
   print("************** THE CONDITION IS NOT VERIFIED ***************"), 
   nl, nl.
    

%%%%%%% search strategies %%%%%%%%%%%%

%%%%%%%%%%%%%%%% exists with loop control %%%%%%%%%%%%%%%%%%%%%%%%%%%%

exists_solve(Node, Cond, Solution) :-
exists([], (nil, Node), Cond, Solution).

exists(Path, (Statement, Node), Cond, [(Statement, Node)|Path]) :- 
exists_final((Statement, Node), Cond).

exists(Path, (Statement, Node), Cond, Solution) :- 
move(Node, Statement1, Node1),
not(member((Statement1, Node1), Path)), 
print("Performed a move from "), 
write(Node), 
print(" to "),
write(Node1),
print(" through the statement "),
write(Statement1),
nl,
nl,
exists([(Statement, Node)|Path], (Statement1, Node1), Cond,  Solution).

%%%%%%%%%%%%%%%% forall %%%%%%%%%%%%%%%%%%%%%%%%%%%%

forall_solve(Node, Cond, Solution) :-
forall([], (nil, Node), Cond, Solution).

forall(Path, (Statement, Node), Cond, [(Statement, Node)|Path]) :- 
forall_final((Statement, Node), Cond).

forall(Path, (Statement, Node), Cond, Solution) :- 
move(Node, Statement1, Node1),
Statement1 \= Cond, 
Node1 \= Cond, 
not(member((Statement1, Node1), Path)), 
print("Performed a move from "), 
write(Node), 
print(" to "),
write(Node1),
print(" through the statement "),
write(Statement1),
nl,
nl,
forall([(Statement, Node)|Path], (Statement1, Node1), Cond,  Solution).



%%%%%%%%%%% Definition of the predicates MOVE and FINAL %%%%%%%%%%%%%%%%%%

move(StateN, Statement, StateM) :-
clause(StateN, Body), 
extract_statement_and_new_state(Body, Statement, StateM).


extract_statement_and_new_state(exec(State), nil, exec(State)) :- true.

extract_statement_and_new_state((Statement, exec(State)), Statement, exec(State)) :- true.

extract_statement_and_new_state((Statement, (!, exec(State))), Statement, exec(State)) :- true.


forall_final((Statement, exec(s(Something, final))), Cond) :-
Cond \= Statement,
Cond \= exec(s(Something, final)), 
print("This forall path terminates unsuccessfully since the state reached, "), write(exec(Something, final)), 
print(", is the final one, and the condition was not met"), nl, nl.

exists_final((_, Cond), Cond):-
print("This exists path terminates successfully meeting the condition because the state reached, "), write(Cond), 
print(", is the right one"), nl, nl.

exists_final((Cond, _), Cond):-
print("This exists path terminates successfully meeting the condition because the statement executed, "), write(Cond), 
print(", is the right one"), nl, nl.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

assert_protocol_clauses([H|T]) :-
   assert(H), 
   assert_protocol_clauses(T).
   
assert_protocol_clauses([]) :- true.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%======================================================================================================================%
% GENERATION OF CODE

generate_reader_code(
   process(
      %%% Definition of the mandatory parameters %%%
      parameters(
         ontology(O), 
         content_language(CL)),
            
      %%% Definition of the protocol name %%%
      protocol_name(ProtName),
      
      %%% Definition of the agent publisher %%%
      agent_publisher(short_name(PSN), long_name(PLN)),
      
      %%% Definition of the agent reader %%%
      agent_reader(short_name(RSN), long_name(RLN)),
     
      %%% Definition of the body contained in the WS_BPEL file %%%
      main_fragment(Body)
      ), 
      
      Program, 
      
      ProtName
) :-
print(
"
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%   STAGE 2: translation from the intermediate XML code to Prolog   %%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   Trace explaining the performed translation   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


"), 
 
generate_body(s(ProtName, 0), s(ProtName, final), PSN, Body, NestedProgram),
flatten(NestedProgram, ProgramWithoutEnd),
append(ProgramWithoutEnd, [exec(s(ProtName, final)) :- true],  Program),
print(
"


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% Generated Prolog program; indentation of clauses is driven by nesting in the original protocol %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

"),  
print_program(Program).

print_program([]) :- true.

print_program([Head|Tail]):-
extract_state_from_clause(Head, State),!,
pretty_print_clause(State, Head),
nl,nl, 
print_program(Tail).


print_program([Head|Tail]):-
print_program(Head),
nl,nl, 
print_program(Tail).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%% Definition of generate_body with final state to reach %%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% generate_body(
%%%            CurrentState,
%%%            StateToReach,
%%%            AtomicBPELActivity,
%%%            GeneratedPrologClause)

generate_body(
   s(Something, Index),
   SFinal, 
   Publisher,
   send(msg(Performative, Content)), 
   [exec(s(Something, Index)) :- (traceable_send(Performative, Content, Publisher), exec(SFinal))]) 
   :-  
   debug(send,  
      s(Something, Index), SFinal, 
      send(msg(Performative, Content)), 
     [exec(s(Something, Index)) :- (traceable_send(Performative, Content, Publisher), exec(SFinal))]).

generate_body(
   s(Something, Index), 
   SFinal,
   Publisher,
   receive(msg(Performative, Content)), 
   [exec(s(Something, Index)) :- 
            (traceable_receive(Performative, Content, Publisher),              
             exec(SFinal))]) 
   :-    
   debug(receive, 
      s(Something, Index),SFinal,
      receive(msg(Performative, Content)), 
      [exec(s(Something, Index)) :- (traceable_receive(Performative, Content, Publisher), 
             exec(SFinal))]).
   
   
generate_body(
   s(Something, Index), 
   SFinal,
   Publisher,
   while(Guard, While_activities),
   [ exec(s(Something, Index)) :- (check_while_guard_passed(s(Something, Index), Guard), !, exec(s(s(Something, Index), 0))) | 
     [ exec(s(Something, Index)) :- exec(SFinal) |
       [ exec(s(Something, Index1)) :- exec(s(Something, Index)) | While_code ]]] 
) :- 
   Index1 is Index + 1,
   debug(while, 
      s(Something, Index), SFinal, 
      while(Guard, While_activities),
      [ exec(s(Something, Index)) :- (check_while_guard_passed(s(Something, Index), Guard), !, exec(s(s(Something, Index), 0))) | 
     [ exec(s(Something, Index)) :- exec(SFinal) |
       [ exec(s(Something, Index1)) :- exec(s(Something, Index))]]]),
   generate_body(s(s(Something, Index), 0), s(Something, Index1), Publisher, While_activities, While_code).


generate_body(
   s(Something, Index), 
   SFinal, 
   Publisher,
   switch(Case_activities), 
   Case_code
) :- 
   debug(switch, 
   s(Something, Index), SFinal,
   switch(Case_activities), 
   []),
   generate_case_body(s(s(Something, Index),0), SFinal, Publisher, Case_activities, Case_code).
   
   
   
generate_body(
   s(Something, Index),
   SFinal,
   Publisher,
   if_then(Guard, Then_activities), 
   [ exec(s(Something, Index)) :- (check_if_guard_passed(s(Something, Index), Guard), !,  exec(s(s(Something, Index), 0))) | 
      [ exec(s(Something, Index)) :- exec(SFinal) | Then_code]]
) :- 
   debug(if_then,
    s(Something, Index),SFinal,
    if_then(Guard, Then_activities), 
    [ exec(s(Something, Index)) :- (check_if_guard_passed(s(Something, Index), Guard), !,  exec(s(s(Something, Index), 0))) | 
      [ exec(s(Something, Index)) :- exec(SFinal)]]),
    generate_body(s(s(Something, Index),0), SFinal, Publisher, Then_activities, Then_code).



generate_body(
   s(Something, Index),
   SFinal,
   Publisher,
   seq(Seq_activities), 
   [exec(s(Something, Index)) :- exec(s(s(Something, Index), 0)) | Seq_code]   
   ) :-  
   debug(sequence, s(Something, Index),SFinal,
   seq(Seq_activities), 
   [exec(s(Something, Index)) :- exec(s(s(Something, Index), 0))]),     
   generate_seq_body(
   s(s(Something, Index),0),
   SFinal, 
   Publisher,
   Seq_activities, 
   Seq_code).


generate_seq_body(
   s(Something, Index),
   SFinal, Publisher,
   [H|[]], 
   [H_code|[]]   
   ) :- !,
   generate_body(s(Something, Index), SFinal, Publisher, H, H_code).

generate_seq_body(
   s(Something, Index),
   SFinal, 
   Publisher,
   [H|T], 
   [H_code|Tail_code]   
   ) :- 
   Index1 is Index + 1, 
   generate_body(s(Something, Index), s(Something, Index1), Publisher, H, H_code),
   generate_seq_body(
   s(Something, Index1),
   SFinal, 
   Publisher,
   T, 
   Tail_code   
   ).
   
     
   

%%%%%%%%%%%%%%%%%%% Definition of generate_case_body %%%%%%%%%%%%%%%%%%%%%%%%%%%%%

generate_case_body(_, _, Publisher,  [], []) :- true.

generate_case_body(s(Something, CaseIndex), S1,  Publisher, [case(Guard, Activity)|Other_case_activities], 
[ exec(Something) :- (check_case_guard_passed(s(Something, CaseIndex), Guard), !,  exec(s(Something, CaseIndex))) | 
  Tail_case_code]
) :- 
   generate_body(s(Something, CaseIndex), S1, Publisher, Activity, Case_code),
   CaseIndex1 is CaseIndex + 1, 
   generate_case_body(s(Something, CaseIndex1), S1, Publisher, Other_case_activities, Other_case_code),
   append(Case_code, Other_case_code, Tail_case_code).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


debug(StatementName, State, FinalState, H, []):- !,
 pretty_print_state_nesting(State),print("Translating: "), print(StatementName), nl, 
 pretty_print_state_nesting(State),print("Current state: "), print(State), nl, 
 pretty_print_state_nesting(State),print("State to reach: "), print(FinalState), nl, 
 pretty_print_state_nesting(State),print("Statement under translation: "), print(H),nl, 
 pretty_print_state_nesting(State),print("Statement traslated: "), print("Indented statements below"), nl, nl. 

debug(StatementName, State, FinalState, H, HCode):-

 pretty_print_state_nesting(State),print("Translating: "), print(StatementName), nl, 
 pretty_print_state_nesting(State),print("Current state: "), print(State), nl, 
 pretty_print_state_nesting(State),print("State to reach: "), print(FinalState), nl, 
 pretty_print_state_nesting(State),print("Statement under translation: "), print(H),nl, 
 pretty_print_state_nesting(State),print("Statement traslated: "), print(HCode), nl, nl. 


pretty_print_state_nesting(s(Something, _)) :- !, print("  "), pretty_print_state_nesting(Something).
pretty_print_state_nesting(_) :- true.

extract_state_from_clause(exec(State) :- Body, State) :- !.
extract_state_from_clause(exec(State), State) :- true.



pretty_print_clause(State, exec(State) :- Body):-
pretty_print_state_nesting(State), print("exec(")
                                 , print(State),
                                   print(") :-"),nl,
pretty_print_body(State, Body).


pretty_print_body(State, (FirstAtom, Rest)) :-
!,
pretty_print_state_nesting(State),print("  "), print(FirstAtom),print(","), nl, 
pretty_print_body(State, Rest).

pretty_print_body(State, LastAtom) :-
pretty_print_state_nesting(State),print("  "), print(LastAtom),print("."), nl. 



flatten([], []):-!.

flatten([H|T], F) :- 
   !,
   flatten(H, FH),
   flatten(T, FT),
   append(FH, FT, F).
   
flatten(Elem, [Elem]) :- true.   
%======================================================================================================================%




%======================================================================================================================%
% PREDICATE FOR CHECKING THE SYNTAX CORRECTNESS OF THE INTERMEDIATE PROTOCOL REPRESENTATION

syntax_correctness(
   process(
      %%% Definition of the mandatory parameters %%%
      parameters(
         ontology(O), 
         content_language(CL)),
            
      %%% Definition of the protocol name %%%
      protocol_name(ProtName),
      
      %%% Definition of the agent publisher %%%
      agent_publisher(short_name(PSN), long_name(PLN)),
      
      %%% Definition of the agent reader %%%
      agent_reader(short_name(RSN), long_name(RLN)),
     
      %%% Definition of the body contained in the WS_BPEL file %%%
      main_fragment(Body)
      ), 
      
      yes
) :-
   is_string(O),
   is_string(CL),
   is_string(ProtName),
   is_string(PSN),
   is_string(PLN),
   is_string(RSN),
   is_string(RLN),
   is_body(Body), !.

syntax_correctness(_, no) :- 
   print("The process identified by "), 
   write(ProtName), 
   write(" did not pass the syntax correctness check"), 
   nl, nl.

%%% A well_formed body is defined according to the following inductive definition

is_body(Body) :- 
   is_structured_activity(Body).

is_atomic_activity(send(Message)) :- 
   is_FIPA_ACL_Message(Message).
   
is_atomic_activity(receive(Message)) :- 
   is_FIPA_ACL_Message(Message).      


is_structured_activity(A) :-
   is_atomic_activity(A).

is_structured_activity(while(Guard, While_activities)) :- 
   is_guard(Guard),
   is_structured_activity(While_activities).

is_structured_activity(switch(Case_activities)) :- 
   is_case_activity(Case_activities).

is_structured_activity(switch_otherwise(Case_activities, Otherwise_activities)) :- 
   is_case_activity(Case_activities), 
   is_structured_activity(Otherwise_activities).

is_structured_activity(if_then(Guard, Then_activities)) :- 
   is_guard(Guard),
   is_structured_activity(Then_activities).

is_structured_activity(if_then_else(Guard, Then_activities, Else_activities)) :- 
   is_guard(Guard),
   is_structured_activity(Then_activities),
   is_structured_activity(Else_activities).

is_structured_activity(seq(Seq_activities)) :- 
   is_sequence_activity(Seq_activities).

is_sequence_activity([]) :- true.

is_sequence_activity([Activity|Other_activities]) :-
   is_structured_activity(Activity),
   is_sequence_activity(Other_activities).

is_case_activity([]) :- true.

is_case_activity([case(Guard, Activity)|Other_case_activities]) :- 
   is_guard(Guard),
   is_structured_activity(Activity), 
   is_case_activity(Other_case_activities).

is_FIPA_ACL_Message(msg(Performative, Content)) :- 
   is_FIPA_ACL_performative(Performative),
   is_string(Content).
               
is_FIPA_ACL_performative("ACCEPT-PROPOSAL").
is_FIPA_ACL_performative("AGREE").   
is_FIPA_ACL_performative("CANCEL").   
is_FIPA_ACL_performative("CFP").   
is_FIPA_ACL_performative("CONFIRM").
is_FIPA_ACL_performative("DISCONFIRM"). 
is_FIPA_ACL_performative("FAILURE").    
is_FIPA_ACL_performative("INFORM"). 
is_FIPA_ACL_performative("INFORM-IF").  
is_FIPA_ACL_performative("INFORM-REF"). 
is_FIPA_ACL_performative("NOT-UNDERSTOOD").
is_FIPA_ACL_performative("PROPAGATE").   
is_FIPA_ACL_performative("PROPOSE").  
is_FIPA_ACL_performative("PROXY").  
is_FIPA_ACL_performative("QUERY-IF").    
is_FIPA_ACL_performative("QUERY-REF").   
is_FIPA_ACL_performative("REFUSE").   
is_FIPA_ACL_performative("REJECT-PROPOSAL").
is_FIPA_ACL_performative("REQUEST").   
is_FIPA_ACL_performative("REQUEST-WHEN").  
is_FIPA_ACL_performative("REQUEST-WHENEVER"). 
is_FIPA_ACL_performative("SUBSCRIBE").
is_FIPA_ACL_performative("UNKNOWN").
   
   
is_guard(no_guard) :- true.
is_guard(unsatisfied_guard) :- true.
is_guard((_, _, _)) :- true.

is_string(_) :- true. % String check to be defined %  

%=======================================================================================================================%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% Message reception management %%%%%%%%%%%%%%%%%%%%%%%

local_mailbox([]) :- true.

traceable_send(A, B, C) :- 
   send(A, B, C), !,
   print("Sent message: "), 
   print(A), print("  "), print(B), print("  "), print(C), print("  "), nl, nl.

traceable_receive(A, B, C) :- 
   receive(VA, VB, VC), 
   check_right_message(A, B, C, VA, VB, VC).
   
 
check_right_message(A, B, C, VA, VB, VC) :- 
   ((var(VA) ; var(VB)) ; var(VC)), !, 
   traceable_receive(A, B, C).  
 
check_right_message(A, B, C, VA, VB, VC) :- 
   (((A \= VA) ; (B \= VB)) ; C \= VC), !, 
   traceable_receive(A, B, C).  

check_right_message(A, B, C, VA, VB, VC) :- 
   print("Received message: "), 
   print(VA), print("  "), print(VB), print("  "), print(VC), print("  "), nl, nl.
   
  
check_while_guard_passed(s(Something, Index), Guard) :- 
   is_true(Guard), !, 
   print("Checked while guard: "), 
   print(s(Something, Index)), print("  "), print(Guard), print("  "), nl, nl.

check_if_guard_passed(s(Something, Index), Guard) :- 
   is_true(Guard), !, 
   print("Checked if guard: "), 
   print(s(Something, Index)), print("  "), print(Guard), print("  "), nl, nl.

check_case_guard_passed(s(Something, Index), Guard) :- 
   is_true(Guard), !, 
   print("Checked case guard: "), 
   print(s(Something, Index)), print("  "), print(Guard), print("  "), nl, nl.

is_true(no_guard) :- true.


member_and_remaining(Elem, List, RemainingList) :-
   member(Elem, List), 
   delete(Elem,List,RemainingList).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

look_for_a_new_useful_protocol(

%%%%%%%%% TBD
%%%%%%%%% CUT AND PASTE THE TERM THAT YOU FIND IN THE .FOL FILE GENERATED BY THE AUML2WSBPEL TRANSLATOR HERE! %%%%%%%%

%%%%%%%%% AFTER HAVING INSERTED THE TERM HERE, MAKE SURE YOU INSERT THE CORRECT JADE ADRRESSES OF BOTH THE 
%%%%%%%%% READER AND THE PUBLISHER INSIDE THE TERM. 
%%%%%%%%% YOU MUST EDIT THE SUBTERMS agent_publisher(short_name('publisher@address'),long_name('fs/fruitSeller:fruitSellerClass')),  
%%%%%%%%% AND agent_reader(short_name('reader@address'),long_name('fb/fruitBuyer:fruitBuyerClass'))
%%%%%%%%% AND SUBSTITUTE ONLY THE "address" STRING WITH THE STRING "name-of-your-computer:1099/JADE", WHERE
%%%%%%%%% OBVIOUSLY name-of-your-computer MUST BE ISTANTIATED WITH THE ACTUAL NAME OF YOUR COMPUTER


process(
   parameters(
      ontology('job_search'),
      content_language('first_order_logic')
   ),
   %%% Definition of the protocol name %%%
   protocol_name('sd JobSearch'),
   %%% Definition of the agent publisher %%%
   agent_publisher(short_name('jobsManager@giocas:1099/JADE'),long_name('jm/jobsManager:jobsManagerClass')),
   %%% Definition of the agent reader %%%
   agent_reader(short_name('jobSearcher@giocas:1099/JADE'),long_name('js/jobSearcher:jobSearcherClass')),
   %%% Definition of the body contained in the WS_BPEL file %%%
   main_fragment(
      seq([
      send(msg('REQUEST','available_jobs()')),
      receive(msg('INFORM','availablable_job_number(NumbersOfJobs)')),
         while(no_guard,
            seq([
            receive(msg('CFP','propose_job(job(J))')),
               switch([
                  case(no_guard,
                  send(msg('NOT-UNDERSTOOD','not_understood()')) 
                  ),
                  case(no_guard,
                  send(msg('ACCEPT-PROPOSAL','accept(job(J))')) 
                  ),
                  case(no_guard,
                  send(msg('REJECT-PROPOSAL','reject(job(J))')) 
                  ),
                  case(no_guard,
                     seq([
                     send(msg('REQUEST','more_info(job(J))')),
                     receive(msg('INFORM','more_info(info(I))')),
                        switch([
                           case(no_guard,
                           send(msg('ACCEPT-PROPOSAL','accept(job(J))')) 
                           ),
                           case(no_guard,
                           send(msg('REJECT-PROPOSAL','reject(job(J))')) 
                           ) 
                        ]) 
                     ]) 
                  ) 
               ]),
               if_then(no_guard,
               receive(msg('INFORM','job_contact_info(contacts(ListOfContacts))')) 
               ) 
            ]) 
         ) 
      ]) 
   ))





).
