package it.unige.disi.aumlbpel.transformers;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.apache.crimson.tree.TextNode;
import org.w3c.dom.*;

import java.io.*;
import java.util.Vector;

 
/**
 * @author Giovanni Casella
 * it.unige.disi.aumlbpel.transformers
 * IntermediateXML_TextualNotation.java
 * Questa classe permette di trasformare trasfomare un file xml in formato 
 * IntermediateXML nel formato Textual Notation (un estensione della notazione di Winikoff per diagrammi AUML testuali).
 */

public class IntermediateXML_TextualNotation {
    //Variabile per la stampa formattata
    private int formatting_tab_number=0;
    /**Main
     * 
     * @author Giovanni Casella
     * it.unige.disi.aumlbpel.transformers
     * IntermediateXML_TextualNotation.java
     *
     */
    public static void main(String[] argv) {
        IntermediateXML_TextualNotation trans = new IntermediateXML_TextualNotation();
        
        String nomeFileIntermediateNotationInput = "E:\\EclipseProject\\FromAUMLToBPELExtended\\resource\\testingFile\\fruitSeller.xml";
        String nomeFileTextualNotationOutput= "E:\\EclipseProject\\FromAUMLToBPELExtended\\resource\\testingFile\\fruitSeller.txt";
        
        trans.tranform(nomeFileIntermediateNotationInput,nomeFileTextualNotationOutput);
    }
    /**Permette di trasformare un file xml in formato 
     * IntermediateXML nel formato Textual Notation (un estensione della notazione di Winikoff per diagrammi AUML testuali).
     * Il risultato della trasformazione
     * viene scritto su PrintStream out. 
     * @param out PrintStream su cui viene scritto l'output ovvero il risultato della trasformazione
     * @param datafile file contenente l'xml di origine
     */
    public void tranform(String nomeFileInput,String nomeFileOutput){
        Document document;
        NodeList nList;
        NodeList rootChild;
        Node nd;
        TextNode tnd;
        Node tagRoot;
        //Apertura e creazione File I/O
        File datafile = new File(nomeFileInput);
        File outputFile = new File(nomeFileOutput);
        OutputStream outS;
        PrintStream out=null;
        try {
            outS = new FileOutputStream(outputFile);
            out=new PrintStream(outS);
        } catch (FileNotFoundException e1) {
            // TODO Auto-generated catch block
            e1.printStackTrace();
        }
        
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        Vector nodeVect;
        try {
            DocumentBuilder builder = factory.newDocumentBuilder();
            document = builder.parse(datafile);
            document.normalize();
            //Estraggo la root 
            nList=document.getElementsByTagName("protocol-diagram-intermediate-notation");
            if (nList.getLength()==1)tagRoot=nList.item(0);
            else throw new IllegalArgumentException("Il file xml contiene errori relativi al tag protocol-diagram-intermediate-notation come root");    
            //Estraggo il nome del protocollo 
            rootChild=tagRoot.getChildNodes();
            nodeVect= extractNodebyName(rootChild,"protocol-name");
            String mandatoryParamStr;

            if(nodeVect.size()!=1){
                throw new IllegalArgumentException("Il file xml contiene errori relativi al tag protocol-name");    
            }
            else {
                nd=(Node) nodeVect.get(0); 
                String st = "start " + getNodeText(nd);
                FormattingPrintln(out,st);
            }
            
            //Estraggo i  param
            nodeVect= extractNodebyName(rootChild,"parameters");
             if(nodeVect.size()!=1){
                throw new IllegalArgumentException("Il file xml contiene errori relativi al tag params che deve essere inserito obbligatoriamente.");    
            }
            else {
                nd=(Node) nodeVect.get(0);
                mandatoryParamStr= getNodeText(nd);//((Node)nodeVect.get(0)).getChildNodes().item(0).getNodeValue();
                String mand = extractValueFromMandatoryParam(mandatoryParamStr); 
                FormattingPrintln(out,mand);
            } 
            // Estraggo gli optional param
            nodeVect= extractNodebyName(rootChild,"optional-param");
            int i;
            for(i=0;i<nodeVect.size();i++)
            {
                nd=(Node) nodeVect.get(i);    
                String s=getNodeText(nd);
                s=s.replaceAll("=","");
                s=s.replaceAll("\"","");
                String st= "optional-param " + s;
                FormattingPrintln(out,st);
                
            }   
            
            // Estraggo le informazioni sugli agenti
            nodeVect= extractNodebyName(rootChild,"agent-publisher");
            if(nodeVect.size()!=1){
                throw new IllegalArgumentException("Il file xml contiene errori relativi al tag agent-publisher");    
            }
            else {
                nd=(Node) nodeVect.get(0);    
                String agentStr= getNodeText(nd);
                String publisher = extractValueFromAgentTag(agentStr); 
                String st= "agent-publisher " + publisher;
                FormattingPrintln(out,st);
                
            } 
            nodeVect= extractNodebyName(rootChild,"agent-reader");
            if(nodeVect.size()!=1){
                throw new IllegalArgumentException("Il file xml contiene errori relativi al tag agent-publisher");    
            }
            else {
                nd=(Node) nodeVect.get(0);    
                String agentStr= getNodeText(nd);
                String reader = extractValueFromAgentTag(agentStr); 
                String st= "agent-reader " + reader;
                FormattingPrintln(out,st);
            }
            
            //Estraggo il Main interaction Fragment e i figli
            nodeVect= extractNodebyName(rootChild,"main-interaction-fragment");
            if(nodeVect.size()!=1){
                throw new IllegalArgumentException("Il file xml non contiene interazioni");    
            }
            else {
                nd=(Node) nodeVect.get(0);    
                nList=nd.getChildNodes();
                for(i=0;i<nList.getLength();i++)
                {
                    nd=nList.item(i);
                    //I figli di Main Fragment possono essere solo
                    //<message> o combined-fragment
                    if(nd.getNodeName().equals("message")) processMessage(nd,out);
                    else if(nd.getNodeName().equals("combined-fragment")) processCombinedFragment(nd,out);
                }  
                
                String st= "finish";
                FormattingPrintln(out,st);
            }
            
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
    
    /**Questa funzione processa i CombinedFragment. Un elemento Combined 
     * fragment rappresenta un blocco alternative, loop, etc. e contiene 
     * messaggi o altri combinedFragment nel qual caso viene generata
     * una chiamata ricorsiva.
     * 
     * @param nd Nodo Co,binedFragment
     * @param out Output
     */
    private void processCombinedFragment(Node nd, PrintStream out) {
         NodeList nList;
        Vector nodeVect;
        nList=nd.getChildNodes();
        String type;
        String guard;
        //Estraggo il type del combinedFragment
        nodeVect= extractNodebyName(nList,"type");
        if(nodeVect.size()!=1){
            throw new IllegalArgumentException("Il file xml contiene errori relativi ad un tag combined-fragment(type)");    
        }
        else {
            nd=(Node) nodeVect.get(0);    
            type= getNodeText(nd);
            if(type.equals("alt")) type="alternative";
            if(type.equals("opt")) type="option";
            String st="box " + type;
            FormattingPrintln(out,st);
        }
        //Incremento lo spazio per la formattazione
        formatting_tab_number++;
        
        //Estraggo le guard condition
        nodeVect= extractNodebyName(nList,"guard");
        if(nodeVect.size()!=1){
            throw new IllegalArgumentException("Il file xml contiene errori relativi ad un tag combined-fragment(guard)");    
        }
        else {
            nd=(Node) nodeVect.get(0);    
            guard = getNodeText(nd);
            if (!guard.equals("")) {
                String st2="guard " + guard;
                FormattingPrintln(out,st2);
            }
        }
        
        //Un combined fragment ha come figli una serie di fragment che gestisco di seguiot
        NodeList FragmentChild;
        nodeVect= extractNodebyName(nList,"fragment");
        Node nd2;
        
        for(int i=0;i<nodeVect.size();i++)
        {
            nd=(Node)nodeVect.get(i);
            FragmentChild = nd.getChildNodes();
            //I figli di ciascun fragment possono essere message o altri combined fragment annidati
            for(int j=0;j<FragmentChild.getLength();j++)
            {
                nd2=FragmentChild.item(j);
                //I figli di Fragment possono essere solo
                //<message> o combined-fragment
                if(nd2.getNodeName().equals("message")) processMessage(nd2,out);
                else if(nd2.getNodeName().equals("combined-fragment")) processCombinedFragment(nd2,out);
            }
            //Nel caso sto processando un blocco alternative per ogni alternativa tranne l'ultima inserisco
            //la parola chiave next.
            if(type.equals("alternative")&&i!=nodeVect.size()-1) FormattingPrintln(out,"next");
         } 
        //Chiudo il combinedFragment Processato
        String st2="end " + type;
        FormattingPrintln(out,st2);
        //Decremento la formattazione
        formatting_tab_number--;
        return;
    }
    
    /**
     * Gestisce i tag message 
     * @param nd Nodo message
     * @param out Output
     */
    private void processMessage(Node nd,PrintStream out) {
        
        NodeList nList;
        Vector nodeVect;
        String sender;
        String communicativeAct;
        nList=nd.getChildNodes();
        
        //Estraggo Il sender del messaggio
        nodeVect= extractNodebyName(nList,"sender");
        if(nodeVect.size()!=1){
            throw new IllegalArgumentException("Il file xml contiene errori relativi ad un tag message(sender)");    
        }
        else {
            nd=(Node) nodeVect.get(0);    
            sender= getNodeText(nd);
        }
        //Estraggo il communicativeAct
        nodeVect= extractNodebyName(nList,"communicative-act");
        if(nodeVect.size()!=1){
            throw new IllegalArgumentException("Il file xml contiene errori relativi ad un tag message(communicative-act)");    
        }
        else {
            nd=(Node) nodeVect.get(0);    
            communicativeAct= getNodeText(nd);
        }
        //Estraggo il contenuto
        String contenuto;
        nodeVect= extractNodebyName(nList,"content");
        if(nodeVect.size()!=1){
            contenuto="";}
        else {
            nd=(Node) nodeVect.get(0);    
            contenuto = getNodeText(nd);
        }
        
        //Stampo la riga
        if(sender.equals("reader")) {
            String st = "message r p " + communicativeAct + " " + contenuto;
            FormattingPrintln(out,st);
            
        }
        else if (sender.equals("publisher")){ 
            String st = "message p r " + communicativeAct+ " " + contenuto;
            FormattingPrintln(out,st);
        }
        return;
    }
    /**
     * Un nodo di tipo non-testo ha a sua volta un figlio di tipo testo 
     * contenente il contenuto testuale del nodo. Questa funzione estrae il testo
     * @param nd Nodo
     * @return testo estratto
     */
    private String getNodeText(Node nd){
        String rs="";
        if(nd.getNodeType() == nd.ELEMENT_NODE){
            nd = nd.getFirstChild();
            if(nd!=null) rs = nd.getNodeValue();
            else rs="";
        }
        return rs;
    }
    /** Restituisce un vettore contenente tutti i nodi 
     *  con un certo nome tagName individuati nella nodeList in input
     * @param nl Lista Nodi in cui ricercare 
     * @param tagName Nome del tag da ricercare 
     * @return Vettore con i nodi in nl aventi tagName
     */
    private Vector extractNodebyName (NodeList nl,String tagName){
        Vector result=new Vector();
        Node tmp;
        for(int i=0;i<nl.getLength();i++){
            tmp=nl.item(i);
            if(tmp.getNodeName().equals(tagName)) result.add(tmp);
        }
        return result;
    }
    
    /** I mandatory param di un protocollo di comunicazione sono esperess nell'xml
     *  intermedio secondo la notazione 
     *  <params> <<params>> ontology = "job_search" CL = "first order logic" </params>
     *  
     * @param s contenuto del tag  param 
     * @return stringhe relative all'ontologia ed al content language
     */
    
    private String  extractValueFromMandatoryParam(String s)
    {
         String result="";
        String ontology;
        String cl;
        int start,end;
        int firstValue;
        int secondValue;
        int clPosition;
        firstValue = s.indexOf("=");
        secondValue = s.indexOf("=",firstValue+1);
        clPosition=s.indexOf("CL");
        
        ontology = s.substring(firstValue,clPosition).trim();
        ontology=ontology.replaceAll("=","");
        ontology=ontology.replaceAll("\"","");
        ontology=ontology.replaceAll("\n","");
        ontology=ontology.trim().replaceAll(" ","_");
        
        if(ontology.length()==0)  throw new IllegalArgumentException("Il file xml contiene errori relativi al tag params (ontology)");    
        
        cl = s.substring(secondValue,s.length()).trim();
        cl=cl.replaceAll("=","");
        cl=cl.replaceAll("\"","");
        cl=cl.replaceAll("\n","");
        cl=cl.trim().replaceAll(" ","_");
        if(cl.length()==0)  throw new IllegalArgumentException("Il file xml contiene errori relativi al tag params (content-language)");    
        result=ontology.trim() + " " + cl.trim();
        return result;
        
    }
    
    /**Estrae le informazioni relative agli agenti. Nella notazione xml intermedia sono
     * presenti i tag <agent-publisher> e <agent-reader>
     * del seguente tipo
     * <agent-publisher short-name="p">jm/jobsManager:jobsManagerClass</agent-publisher>
     * La funzione retituisce i dati in un formato proprio del
     * linguaggio testuale
     * @param s Contenuto di <agent-publisher> o <agent-reader>
     * @return informazioni formattate
     */
    private String  extractValueFromAgentTag(String s)
    {
        String result="";
        String name,role,classe;
        int start,end;
        
        end=s.indexOf("/");
        name=s.substring(0,end).trim();
        if(name.length()==0) name="df-agent-name";
        
        start = s.indexOf("/")+1;
        end=s.indexOf(":");
        role=s.substring(start,end).trim();
        if(role.length()==0) name="df-role-name";
        
        start = s.indexOf(":")+1;
        classe=s.substring(start,s.length()).trim();
        if(classe.length()==0) name="df-class-name";
        
        result = name+" "+role+" "+classe+" "; 
        return result;
        
    }
    /**
     * Stampa la stringa s formattandola correttamente
     * @param out
     * @param s
     */
    private void FormattingPrint(PrintStream out,String s){
        for(int i = 0 ; i<formatting_tab_number;i++) s="\t"+s;
        out.print(s);     
        
    }
    /**
     * Stampa la stringa s formattandola correttamente e andando a capo
     * @param out
     * @param s
     */
    private void FormattingPrintln(PrintStream out,String s){
        for(int i = 0 ; i<formatting_tab_number;i++) s="\t"+s;
        out.println(s);     
    }
    
}
