package it.unige.disi.aumlbpel.transformers;

/**
 * @author Giovanni Casella 
 * it.unige.disi.aumlbpel.transformers
 * IntermediateXML_TextualNotation.java 
 * Trasfoma un file xml in formato IntermediateXML nel documento
 * wsdl adatto a supportare un processo bpel.
 */
import org.jdom.*;
import javax.xml.parsers.DocumentBuilderFactory;
import org.jdom.output.*;

import java.io.*;
import java.util.List;
import java.util.Vector;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import java.io.IOException;

public class IntermediateXML_WSDL_BPEL {
    final static String wsdlBaseFileName = "resource\\transformer_stylesheet\\base.wsdl";

    final static String bpelBaseFileName = "resource\\transformer_stylesheet\\base.bpel";

    //ID messaggi
    int mess_index = 1;

    //ID variabili
    int var_index = 1;

    //root dei DOM rappresentanti i relativi Files
    private Element rootWSDL;

    private Element rootBPEL;

    private Element rootXML;

    //  root elementi aggiunti in types esistente del file WSDL
    private Element rootAddedTypesWSDL = new Element("ADDED");

    //root elemento messages file wsdl
    private Element rootMessagesWSDL = new Element("messages");
    //root elemento per la dichiarazione dei contenuti dei vari messaggi
    private Element rootContentsWSDL = new Element("contents");
    //root variabili bpel
    Element rootVariablesBpel = new Element("variables");

    //root del processo bpel vero e proprio
    Element rootProcess = new Element("process");

    //  Ultimo elemento inserito nel processoBpel
    private Element currentCFBpelInserted;

    //root delle due port type
    private Element rootPubPortType;

    private Element rootReadPortType;

    private String pubRole;

    private String readRole;

    public static void main(String[] argv) {
        IntermediateXML_WSDL_BPEL trans = new IntermediateXML_WSDL_BPEL();

        String nomeFileIntermediateNotationInput = "E:\\EclipseProject\\FromAUMLToBPELExtended\\resource\\testingFile\\fruitSeller.xml";
        String nomeFileWsdlOutput = "E:\\EclipseProject\\FromAUMLToBPELExtended\\resource\\testingFile\\fruitSeller.wsdl";
        String NomeFileBpel = "E:\\EclipseProject\\FromAUMLToBPELExtended\\resource\\testingFile\\fruitSeller.bpel";
        
        trans.tranformXMLtoWSDL_BPEL(nomeFileIntermediateNotationInput,
                nomeFileWsdlOutput, NomeFileBpel);
    }

    /**
     * Genera i documenti WSDL e Bpel a partire dall'Xml in input Il file XML
     * deve contenere codice XML Intermedio
     * 
     * @param nomeFileInput
     * @param nomeFileOutput
     */
    public void tranformXMLtoWSDL_BPEL(String nomeFileXmlInput,
            String nomeWsdlOutput, String nomeBpelOutput) {
        Vector elementVect;
        //Apertura e creazione File I/O
        initializeDOM(nomeFileXmlInput);
        createDOMS();
        attachAddedTypesWSDL();
        writeDoms(nomeWsdlOutput, nomeBpelOutput);
    }

    /**
     * Scrive sui rispettivi File l'albero DOM rapprensetante il documento Bpel
     * e quello che rapprensenta il documento Wsdl
     * 
     * @param nomeFileWsdlOutput
     * @param nomeFileBpelOutput
     */
    private void writeDoms(String nomeFileWsdlOutput, String nomeFileBpelOutput) {
        File outputWsdlFile = new File(nomeFileWsdlOutput);
        File outputBpelFile = new File(nomeFileBpelOutput);
        OutputStream outWsdl, outBpel;
        try {
            outWsdl = new FileOutputStream(outputWsdlFile);
            outBpel = new FileOutputStream(outputBpelFile);
            //          SCRITTURA OUTPUT Wsdl e Bpel
            XMLOutputter outputter = new XMLOutputter();
            outputter.output(rootWSDL, outWsdl);
            outputter.output(rootBPEL, outBpel);
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }

    }

    /**
     * Inizializza gli alberi DOM (xml,wsdl e bpel) in base ai parametri di
     * input ed ai file template base.wsdl e base.bpel che si trovano in
     * resource/sylesheet
     * 
     * @param nomeFileXmlInput
     * @param nomeWsdlOutput
     * @param nomeBpelOutput
     */
    private void initializeDOM(String nomeFileXmlInput) {
        File inputfile = new File(nomeFileXmlInput);
        File wsdlBase = new File(wsdlBaseFileName);
        File bpelBase = new File(bpelBaseFileName);
        //Inizilaizzazione gestori DOM
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        Vector nodeVect;
        SAXBuilder builder = new SAXBuilder();
        String namespace_wsdl_bpel = "http://www.disi.unige.it/auml_bpel/wsdl/" + nomeFileXmlInput;

        try {
            //      DOM di Input
            Document xmlDocument;
            xmlDocument = builder.build(nomeFileXmlInput);
            rootXML = xmlDocument.getRootElement();
            // APERUTURA FILE DOM WSDL E BPEL DI BASE E CREAZIONE DOM
            Document wsdlDocument = builder.build(wsdlBase);
            rootWSDL = wsdlDocument.getRootElement();
            rootWSDL.setAttribute("targetNamespace",namespace_wsdl_bpel);

            
            
            Document bpelDocument = builder.build(bpelBase);
            rootBPEL = bpelDocument.getRootElement();
            Namespace ns = Namespace.getNamespace("lns",namespace_wsdl_bpel);
   
            rootBPEL.addNamespaceDeclaration(ns);
            currentCFBpelInserted = rootProcess;
            rootBPEL.setAttribute("targetNamespace","http://www.disi.unige.it/auml_bpel/bpel/"+nomeFileXmlInput);

            
        } catch (JDOMException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

    }

    /**
     * Riempie i DOM WSDL E BPEL estraendo i dati dal DOM XML
     * 
     * @param rootXml
     * @param rootWsdl
     * @param rootBpel
     */
    private void createDOMS() {
        Vector elementVect;
        Element tmp;
        Element t;
        addProtocolElementsDeclaration();
        //Dichirazione messaggi nel file wsdl
        
        Element messages = 
        //I messaggi li attacco tutti alla fine
        rootMessagesWSDL = new Element("messages", rootWSDL.getNamespace());
        rootContentsWSDL = new Element("contents", rootWSDL.getNamespace());
        rootReadPortType = new Element("portType", rootWSDL.getNamespace());
        rootReadPortType.setAttribute("name", "readerPT");
        rootPubPortType = new Element("portType", rootWSDL.getNamespace());
        rootPubPortType.setAttribute("name", "publisherPT");
        
        manageProcess();
        //Aggiungo la dichiarazione dei messaggi trovati nel processo nel WSDL
        attachContentsElementWSDL();
        attachMessagesWSDL();
        attachVariablesBPEL();
        attachPortTypes();
        addPartnerLinkTypeDeclaration();
        attachProcess();

    }

    /**
     * 
     */
    private void attachContentsElementWSDL() {
        	
            List l;
            l = rootContentsWSDL.getChildren();
            while (l.size() > 0) {
                Element t = (Element) l.get(0);
                //stacco l'elemento dal vecchio padre altrimenti da errore
                //facendo il detach lista size si decrementa automaticamente
                //e quindi prendo sempre l'eleemnto in testa.
                t.detach();
                rootAddedTypesWSDL.addContent(t);
               
            }

        
        
    }

    private void attachVariablesBPEL() {
        rootVariablesBpel.setNamespace(rootBPEL.getNamespace());
        rootBPEL.addContent(rootVariablesBpel);
    }

    /**
     * Dopo aver visionato il processo inserisce la dichiarazione di tutti i
     * messaggi
     *  
     */
    private void attachPortTypes() {

        rootAddedTypesWSDL.addContent(rootPubPortType);
        rootAddedTypesWSDL.addContent(rootReadPortType);

    }

    /**
     * Dopo aver visionato il processo inserisce la dichiarazione di tutti i
     * messaggi trovati
     *  
     */
    private void attachMessagesWSDL() {
        List l;
        l = rootMessagesWSDL.getChildren();
        while (l.size() > 0) {
            Element t = (Element) l.get(0);
            //stacco l'elemento dal vecchio padre altrimenti da errore
            //facendo il detach lista size si decrementa automaticamente
            //e quindi prendo sempre l'eleemnto in testa.
            t.detach();
            rootAddedTypesWSDL.addContent(t);
        }

    }

    private void attachProcess() {

        List l;
        l = rootProcess.getChildren();
        while (l.size() > 0) {
            Element t = (Element) l.get(0);
            //stacco l'elemento dal vecchio padre altrimenti da errore
            //facendo il detach lista size si decrementa automaticamente
            //e quindi prendo sempre l'eleemnto in testa.
            t.detach();
            rootBPEL.addContent(t);
        }

    }

    private void attachAddedTypesWSDL() {

        Element types;
        Vector v = extractNodebyName(rootWSDL.getChildren(), "types");
        types = (Element) v.get(0);

        List l;
        l = rootAddedTypesWSDL.getChildren();
        while (l.size() > 0) {
            Element t = (Element) l.get(0);
            //stacco l'elemento dal vecchio padre altrimenti da errore
            //facendo il detach lista size si decrementa automaticamente
            //e quindi prendo sempre l'eleemnto in testa.
            t.detach();
            types.addContent(t);
        }
    }

    /**
     * Analizza il processo descritto a partire dal tag
     * main-interaction-fragment e aggiorna di conseguenza il dom WSDL ed il dom
     * BPEL
     *  
     */
    private void manageProcess() {
        Vector nodeVect;
        List children;
        nodeVect = extractNodebyName(rootXML.getChildren(),
                "main-interaction-fragment");
        Element tmp = (Element) nodeVect.get(0);
        children = tmp.getChildren();
        Element t;
        for (int i = 0; i < children.size(); i++) {
            t = (Element) children.get(i);
            if (t.getName().equals("message")) {
                processMessage(t);
            } else if (t.getName().equals("combined-fragment"))
                processCombinedFragment(t);

        }

    }

    private void addMessageAndOperationWsdlDec(String sender,
            String communicative_act, String ContenutoMessaggio) {
        //Si presuppone che nello schema wsdl l'elemento denominato
        //confirm_ca sia adatto a rappresentare in communicative act confirm e
        // cosi via
        //quind x_ca rappresenta il communicative act c.
        String SenderType;
        communicative_act = communicative_act + "_ca";
        String messageName = "Mess_" + mess_index;
        
        
        Element message = new Element("message", rootMessagesWSDL
                .getNamespace());
        message.setAttribute("name", messageName);

        Element partPerf = new Element("part", message.getNamespace());
        partPerf.setAttribute("name", "performative");
        partPerf.setAttribute("element", communicative_act);
        message.addContent(partPerf);

        Element partPartecipant = new Element("part", message.getNamespace());
        partPartecipant.setAttribute("name", "Participant");
        
        if (sender.equals("reader"))
            SenderType = "tns:Participant_MsgFromReader";
        else
            SenderType = "tns:Participant_MsgFromPublisher";
        
        partPartecipant.setAttribute("type", SenderType);
        message.addContent(partPartecipant);
        String messageContentElement = "content_of_"+messageName;
        addIndipendentParts(message,messageContentElement);
        
        rootMessagesWSDL.addContent(message);
        
        //Creo il content adatto per il messaggio appena definito
        Namespace ns = Namespace.getNamespace("xs",
        "http://www.w3.org/2001/XMLSchema");
        Element messageContent = new Element("element", ns);
        messageContent.setAttribute("name", messageContentElement);
        messageContent.setAttribute("type", "xs:string");
        messageContent.setAttribute("fixed", ContenutoMessaggio);
        rootContentsWSDL.addContent(messageContent);
        
        
        //Creo l'operation nella port_type wsdl del Publisher
        String name;
        Element op = new Element("operation", rootWSDL.getNamespace());
        //Messaggio inviato dal reder e ricevuto dal publisher
        if (sender.equals("reader")) {
            name = "RCV_" + messageName;
            op.setAttribute("name", name);
            Element input = new Element("input", op.getNamespace());
            input.setAttribute("message", "tns:" + messageName);
            op.addContent(input);
        } else {//messaggio inviato dal publisher e ricevuto dal reader
            name = "SND_" + messageName;
            op.setAttribute("name", name);
            Element input = new Element("output", op.getNamespace());
            input.setAttribute("message", "tns:" + messageName);
            op.addContent(input);
        }
        rootPubPortType.addContent(op);

        //      Creo l'operation nella port_type wsdl del Reader
        op = new Element("operation", rootWSDL.getNamespace());
        //Messaggio inviato dal reder e ricevuto dal publisher
        if (sender.equals("reader")) {
            name = "SND_" + messageName;
            op.setAttribute("name", name);
            Element input = new Element("output", op.getNamespace());
            input.setAttribute("message", "tns:" + messageName);
            op.addContent(input);
        } else {//messaggio inviato dal publisher e ricevuto dal reader
            name = "RCV_" + messageName;
            op.setAttribute("name", name);
            Element input = new Element("input", op.getNamespace());
            input.setAttribute("message", "tns:" + messageName);
            op.addContent(input);
        }
        rootReadPortType.addContent(op);

    }

    private void addReceiveSendBpel(String sender, String communicative_act) {
        //Controllo se e' il primo messaggio, cio quello che avvia il
        // protocollo
        if (mess_index == 1) {
            processFirstMessage(sender, communicative_act);
            return;
        }

        if (sender.equals("reader")) {
           
            Element send = new Element("invoke", rootBPEL.getNamespace());
            send.setAttribute("partnerLink", "readerPL");
            send.setAttribute("portType", "lns:readerPT");
            send.setAttribute("operation", "SND_" + "Mess_" + mess_index);
            currentCFBpelInserted.addContent(send);
            
            Element receive = new Element("receive", rootBPEL.getNamespace());
            receive.setAttribute("partnerLink", "publisherPL");
            receive.setAttribute("portType", "lns:publisherPT");
            receive.setAttribute("operation", "RCV_" + "Mess_" + mess_index);
            currentCFBpelInserted.addContent(receive);

        } else if (sender.equals("publisher")) {
            Element send = new Element("invoke", rootBPEL.getNamespace());
            send.setAttribute("partnerLink", "publisherPL");
            send.setAttribute("portType", "lns:publisherPT");
            send.setAttribute("operation", "SND_" + "Mess_" + mess_index);
            currentCFBpelInserted.addContent(send);
            
            Element receive = new Element("receive", rootBPEL.getNamespace());
            receive.setAttribute("partnerLink", "readerPL");
            receive.setAttribute("portType", "lns:readerPT");
            receive.setAttribute("operation", "RCV_" + "Mess_" + mess_index);
            currentCFBpelInserted.addContent(receive);

            
        }

    }

    private void processFirstMessage(String sender, String communicative_act) {
        if (!sender.equals("reader"))
            throw new IllegalArgumentException(
                    "Il protocollo pu iniziare esclusivamente con un msssagggio inviato dal reader");
        Element receive = new Element("receive", rootBPEL.getNamespace());
        /*
         * Controllo se il protocollo inzia con un messaggio. In questo caso
         * inserisco un blocco seqence.
         */
        if (currentCFBpelInserted == rootProcess)
            insertFirstSequence();

        receive.setAttribute("partnerLink", "publisherPL");
        receive.setAttribute("portType", "lns:publisherPT");
        receive.setAttribute("operation", "RCV_" + "Mess_" + mess_index);
        receive.setAttribute("createIstance", "yes");
        currentCFBpelInserted.addContent(receive);

        //Element send = new Element("invoke", rootBPEL.getNamespace());
        //send.setAttribute("partnerLink", "readerPL");
        //send.setAttribute("portType", "lns:readerPT");
        //send.setAttribute("operation", "SND_" + "Mess_" + mess_index);
        //currentCFBpelInserted.addContent(send);

    }

    private void processMessage(Element t) {

        String SenderType;
        String sender;
        String communicative_act;
        String content;
        Vector v;
        v = extractNodebyName(t.getChildren(), "sender");
        sender = ((Element) v.get(0)).getText();
        v = extractNodebyName(t.getChildren(), "communicative-act");
        communicative_act = ((Element) v.get(0)).getText();
        v = extractNodebyName(t.getChildren(), "content");
        content = ((Element) v.get(0)).getText();
        
        
        addMessageAndOperationWsdlDec(sender, communicative_act,content);
        addReceiveSendBpel(sender, communicative_act);
        mess_index++;

    }

    /**
     * Aggiunge alla dichiarazione dei messaggi Wsdl le seguenti parti sempre
     * uguali
     * 
     * <part name="Content" type="tns:message_content"/> <part
     * name="Content_Language" element="tns:content_language_name"/> <part
     * name="Ontology" element="tns:ontology_name"/> <part name="Protocol"
     * element="tns:protcol_name"/>
     * 
     * @author Giovanni Casella it.unige.disi.aumlbpel.transformers
     *         IntermediateXML_WSDL.java
     *  
     */
    private void addIndipendentParts(Element message, String messageCOntentElement) {
        Element partContent = new Element("part", message.getNamespace());
        partContent.setAttribute("name", "Content");
        String messageContent = "tns:"+messageCOntentElement;
        partContent.setAttribute("element", messageContent);
        message.addContent(partContent);

        Element Content_Language = new Element("part", message.getNamespace());
        Content_Language.setAttribute("name", "Content_Language");
        Content_Language.setAttribute("element", "tns:content_language_name");
        message.addContent(Content_Language);

        Element Content_Ontology = new Element("part", message.getNamespace());
        Content_Ontology.setAttribute("name", "Ontology");
        Content_Ontology.setAttribute("element", "tns:ontology_name");
        message.addContent(Content_Ontology);

        Element Protocol = new Element("part", message.getNamespace());
        Protocol.setAttribute("name", "Protocol");
        Protocol.setAttribute("element", "tns:protocol_name");
        message.addContent(Protocol);

    }

    private void processCombinedFragment(Element t) {

        Vector v = extractNodebyName(t.getChildren(), "type");
        Element type = (Element) v.get(0);
        String ti = type.getText();
        if (ti.equals("loop"))
            processLoopCF(t);
        else if (ti.equals("seq"))
            processSequenceCF(t);
        else if (ti.equals("alt"))
            processAltCF(t);
        else if (ti.equals("opt"))
            processOptCF(t);

    }

    private void insertFirstSequence() {

        Element tmp;

        Element seq = new Element("sequence", rootBPEL.getNamespace());
        currentCFBpelInserted.addContent(seq);
        currentCFBpelInserted = seq;

    }

    private void processSequenceCF(Element cf) {

        Element tmp;

        Element seq = new Element("sequence", rootBPEL.getNamespace());
        currentCFBpelInserted.addContent(seq);
        currentCFBpelInserted = seq;
        Vector v = extractNodebyName(cf.getChildren(), "fragment");
        for (int i = 0; i < v.size(); i++) {

            Element frag = (Element) v.get(i);
            List frag_childs = frag.getChildren();
            for (int h = 0; h < frag_childs.size(); h++) {
                Element t2 = (Element) frag_childs.get(h);
                if (t2.getName().equals("message")) {
                    processMessage(t2);

                } else if (t2.getName().equals("combined-fragment")) {
                    processCombinedFragment(t2);
                }

                currentCFBpelInserted = seq;
            }
        }

    }

   

    private void processLoopCF(Element cf) {
        Vector v = extractNodebyName(cf.getChildren(), "fragment");
        Element t, tmp;
        List FragmentChilds;
        //Dichiarazione variabile del Loop nei types WSDL
        String name = "continue_" + var_index + "_type";
        Namespace ns = Namespace.getNamespace("xs",
                "http://www.w3.org/2001/XMLSchema");
        Element el = new Element("element", ns);
        el.setAttribute("name", name);
        Element complexType = new Element("complexType", ns);
        Element sequence = new Element("sequence", ns);
        Element internalEl = new Element("element", ns);
        internalEl.setAttribute("name", "value");
        internalEl.setAttribute("type", "xs:boolean");
        internalEl.setAttribute("form", "qualified");
        sequence.addContent(internalEl);
        complexType.addContent(sequence);
        el.addContent(complexType);
        rootAddedTypesWSDL.addContent(el);

        //Dichiarazione variabile BPEL
        rootVariablesBpel.setNamespace(rootBPEL.getNamespace());
        name = "continue_" + var_index;
        Element v2 = new Element("variable", rootBPEL.getNamespace());
        v2.setAttribute("name", name);
        v2.setAttribute("element", "lns:" + name + "_type");
        rootVariablesBpel.addContent(v2);

        //assegnazione copy di valore opaco
        Element copy = new Element("copy", rootBPEL.getNamespace());
        Element from = new Element("from", rootBPEL.getNamespace());
        from.setAttribute("opaque", "yes");
        copy.addContent(from);
        Element to = new Element("to", rootBPEL.getNamespace());
        String to_val = "$" + "continue_" + var_index + ".value";
        to.setText(to_val);
        copy.addContent(to);
        currentCFBpelInserted.addContent(copy);

        Element while_el = new Element("while", rootBPEL.getNamespace());
        while_el.setAttribute("condition", "$" + "continue_" + var_index
                + ".value" + "=true");
        currentCFBpelInserted.addContent(while_el);
        currentCFBpelInserted = while_el;

        var_index++;
        for (int i = 0; i < v.size(); i++) {
            t = (Element) v.get(i);
            FragmentChilds = t.getChildren();
            for (int j = 0; j < FragmentChilds.size(); j++) {
                tmp = (Element) FragmentChilds.get(j);
                if (tmp.getName().equals("message")) {
                    processMessage(tmp);

                } else if (tmp.getName().equals("combined-fragment"))
                    processCombinedFragment(tmp);
            }
            currentCFBpelInserted = while_el;
        }

        //Aggiorno la variabile continue dopo ogni ciclo loop
        //assegnazione copy di valore opaco
        Element copy2 = new Element("copy", rootBPEL.getNamespace());
        Element from2 = new Element("from", rootBPEL.getNamespace());
        from2.setAttribute("opaque", "yes");
        copy2.addContent(from2);
        Element to2 = new Element("to", rootBPEL.getNamespace());
        to2.setText(to_val);
        copy2.addContent(to2);
        currentCFBpelInserted.addContent(copy2);

    }

    private void processAltCF(Element cf) {
        Vector fragment_vector = extractNodebyName(cf.getChildren(), "fragment");
        Element t, tmp;
        List FragmentChilds;
        int alternative_case = fragment_vector.size();

        //Dichiarazione variabile dell'Alt nel WSDL
        String name = "choose_" + var_index + "_type";
        Namespace ns = Namespace.getNamespace("xs",
                "http://www.w3.org/2001/XMLSchema");
        Element el = new Element("element", ns);
        el.setAttribute("name", name);
        Element simpleType = new Element("simpleType", ns);
        el.addContent(simpleType);
        Element int_restriction = new Element("restriction", ns);
        simpleType.addContent(int_restriction);
        int_restriction.setAttribute("base", "xsd:integer");
        Element minInclusive = new Element("minInclusive", ns);
        String min = "1";
        minInclusive.setAttribute("value", min);
        int_restriction.addContent(minInclusive);
        Element maxInclusive = new Element("maxInclusive", ns);
        maxInclusive.setAttribute("value", fragment_vector.size() + "");
        int_restriction.addContent(maxInclusive);
        rootAddedTypesWSDL.addContent(el);
        
        //Dichiarazione variabile bpel
        name = "choose_" + var_index;
        Element v1 = new Element("variable", rootBPEL.getNamespace());
        v1.setAttribute("name", name);
        v1.setAttribute("element", "lns:" + name + "_type");
        rootVariablesBpel.addContent(v1);

        //assegnazione copy di valore opaco
        Element copy = new Element("copy", rootBPEL.getNamespace());
        Element from = new Element("from", rootBPEL.getNamespace());
        from.setAttribute("opaque", "yes");
        copy.addContent(from);
        Element to = new Element("to", rootBPEL.getNamespace());
        to.setText("$" + "choose" + var_index + ".value");
        copy.addContent(to);
        currentCFBpelInserted.addContent(copy);

        Element switch_el = new Element("switch", rootBPEL.getNamespace());
        currentCFBpelInserted.addContent(switch_el);
        currentCFBpelInserted = switch_el;

        Element caso_switch;
        String choose_val = "$" + "choose" + var_index;
        var_index++;
        for (int h = 0; h < fragment_vector.size(); h++) {
            caso_switch = (Element) fragment_vector.get(h);

            Element case_el = new Element("case", rootBPEL.getNamespace());
            String c = choose_val + ".value=" + (h + 1);
            case_el.setAttribute("condition", c);
            currentCFBpelInserted.addContent(case_el);
            currentCFBpelInserted = case_el;
            List childs = caso_switch.getChildren();
            for (int i = 0; i < childs.size(); i++) {
                tmp = (Element) childs.get(i);
                if (tmp.getName().equals("message")) {
                    processMessage(tmp);

                } else if (tmp.getName().equals("combined-fragment")) {
                    processCombinedFragment(tmp);
                }
                currentCFBpelInserted = case_el;
            }
            currentCFBpelInserted = switch_el;
        }

        currentCFBpelInserted = switch_el;
    }

    private void processOptCF(Element cf) {
        Vector v = extractNodebyName(cf.getChildren(), "fragment");
        Element t, tmp;
        List FragmentChilds;

        //Dichiarazione variabile condition nei types WSDL
        String name = "condition_" + var_index + "_type";
        Namespace ns = Namespace.getNamespace("xs",
                "http://www.w3.org/2001/XMLSchema");
        Element el = new Element("element", ns);
        el.setAttribute("name", name);
        Element complexType = new Element("complexType", ns);
        Element sequence = new Element("sequence", ns);
        Element internalEl = new Element("element", ns);
        internalEl.setAttribute("name", "value");
        internalEl.setAttribute("type", "xs:boolean");
        internalEl.setAttribute("form", "qualified");
        sequence.addContent(internalEl);
        complexType.addContent(sequence);
        el.addContent(complexType);
        rootAddedTypesWSDL.addContent(el);

        //Dichiarazione variabile BPEL
        rootVariablesBpel.setNamespace(rootBPEL.getNamespace());
        name = "condition_" + var_index;
        Element v2 = new Element("variable", rootBPEL.getNamespace());
        v2.setAttribute("name", name);
        v2.setAttribute("element", "lns:" + name + "_type");
        rootVariablesBpel.addContent(v2);

        //assegnazione copy di valore opaco
        Element copy = new Element("copy", rootBPEL.getNamespace());
        Element from = new Element("from", rootBPEL.getNamespace());
        from.setAttribute("opaque", "yes");
        copy.addContent(from);
        Element to = new Element("to", rootBPEL.getNamespace());
        to.setText("$" + "condition_" + var_index + ".value");
        copy.addContent(to);
        currentCFBpelInserted.addContent(copy);
        

        Element if_el = new Element("if", rootBPEL.getNamespace());
        Element condition = new Element("condition", rootBPEL.getNamespace());
        condition.setAttribute("condition", "$" + "condition_" + var_index
                + ".value" + "=true");
        if_el.addContent(condition);
        currentCFBpelInserted.addContent(if_el);
        currentCFBpelInserted = if_el;
        var_index++;

        Element then_el = new Element("then", rootBPEL.getNamespace());
        if_el.addContent(then_el);
        currentCFBpelInserted = then_el;

        //Element sequence_el=new Element("sequence",rootBPEL.getNamespace());
        //then_el.addContent(sequence_el);

        for (int i = 0; i < v.size(); i++) {
            t = (Element) v.get(i);
            FragmentChilds = t.getChildren();
            for (int j = 0; j < FragmentChilds.size(); j++) {
                tmp = (Element) FragmentChilds.get(j);
                if (tmp.getName().equals("message")) {
                    processMessage(tmp);

                } else if (tmp.getName().equals("combined-fragment")) {
                    processCombinedFragment(tmp);
                }
                currentCFBpelInserted = if_el;
            }
            currentCFBpelInserted = if_el;
        }

    }

    /**
     * Inserisce nel file WSDL le dichiarazioni relativi a nome e ruolo degli
     * agenti Richiama la funzione che aggiunge nel file BPEL le dichiarazioni
     * relative ai partner Link.
     * 
     * @param types
     * @param xmlRoot
     * @param rootBpel
     */
    private void addProtocolElementsDeclaration() {
        Element tmp, nuovo;
        Vector elementVector;
        String pubName, pubRole, readName, readRole;
        Namespace ns = Namespace.getNamespace("xs",
                "http://www.w3.org/2001/XMLSchema");
        elementVector = extractNodebyName(rootXML.getChildren(), "parameters");

        if (elementVector.size() == 0)
            throw new IllegalArgumentException(
                    " Il file xml contiene errori relativi al tag params che deve essere inserito obbligatoriamente.");

        tmp = (Element) elementVector.get(0);
        String res[] = extractValueFromMandatoryParam(tmp.getText());
        String ontology = res[0];
        String content_language_name = res[1];
        //ELEMENTI SCHEMA WSDL
        addElement(rootAddedTypesWSDL, ns, "ontology_name", "xs:string",
                ontology);
        addElement(rootAddedTypesWSDL, ns, "content_language_name",
                "xs:string", content_language_name);
        //Protocol Name
        elementVector = extractNodebyName(rootXML.getChildren(),
                "protocol-name");
        tmp = (Element) elementVector.get(0);
        addElement(rootAddedTypesWSDL, ns, "protocol_name", "xs:string", tmp
                .getText());
        //Agent Publisher
        elementVector = extractNodebyName(rootXML.getChildren(),
                "agent-publisher");
        tmp = (Element) elementVector.get(0);
        String[] val = ExtractNameRoleClass(tmp.getText(), true);
        pubName = val[0];
        pubRole = val[1];
        addElement(rootAddedTypesWSDL, ns, "publisher_name", "xs:string",
                val[0]);
        addElement(rootAddedTypesWSDL, ns, "publisher_role", "xs:string",
                val[1]);
        //Agent Reader
        elementVector = extractNodebyName(rootXML.getChildren(), "agent-reader");
        tmp = (Element) elementVector.get(0);
        val = ExtractNameRoleClass(tmp.getText(), false);
        readName = val[0];
        readRole = val[1];
        nuovo = new Element("element", ns);
        addElement(rootAddedTypesWSDL, ns, "reader_name", "xs:string", val[0]);
        addElement(rootAddedTypesWSDL, ns, "reader_role", "xs:string", val[1]);
        //Partecipanti dei messaggi dal publisher e dal reader
        addMessageParticipiantWsdl();
        //PARTNERLINK BPEL
        addPartnerLinkDeclaration(pubRole, readRole);

        //VARIABILI BPEL E WSDL
        /*
         * Element var = new Element("variables");
         * var.setNamespace(rootXML.getNamespace()); rootWSDL = var; Element
         * var2 = new Element("variables");
         * var.setNamespace(rootXML.getNamespace()); rootBPEL.addContent(var2);
         * rootBPEL= var2;
         */
    }

    /**
     * Crea le seguenti dicihiarazioni nel file WSDL <xs:element
     * name="Partecipant_MsgFromPublisher"> <xs:complexType><xs:sequence>
     * <xs:element name="sender" type="tns:publisher_name"/> <xs:element
     * name="receiver" type="tns:reader_name"/> <xs:element name="reply_to"
     * type="tns:publisher_name"/> </xs:sequence> </xs:complexType>
     * </xs:element> <xs:element name="Partecipant_MsgFromReader">
     * <xs:complexType><xs:sequence><xs:element name="sender"
     * type="tns:reader_name"/> <xs:element name="receiver"
     * type="tns:publisher_name"/> <xs:element name="reply_to"
     * type="tns:reader_name"/> </xs:sequence> </xs:complexType> </xs:element>
     * </types>
     */
    private void addMessageParticipiantWsdl() {
        //Mittente Publisher
        Namespace ns = Namespace.getNamespace("xs",
                "http://www.w3.org/2001/XMLSchema");

        Element element = new Element("element", ns);
        element.setAttribute("name", "Participant_MsgFromPublisher");

        Element complexType = new Element("complexType", ns);
        element.addContent(complexType);
        Element sequence = new Element("sequence", ns);
        complexType.addContent(sequence);
        Element sender = new Element("element", ns);
        sender.setAttribute("name", "sender");
        sender.setAttribute("type", "tns:publisher_name");
        sequence.addContent(sender);
        Element receiver = new Element("element", ns);
        receiver.setAttribute("name", "receiver");
        receiver.setAttribute("type", "tns:reader_name");
        sequence.addContent(receiver);
        Element reply_to = new Element("element", ns);
        reply_to.setAttribute("name", "reply_to");
        reply_to.setAttribute("type", "tns:publisher_name");
        sequence.addContent(reply_to);
        rootAddedTypesWSDL.addContent(element);
        //  Mittente Reader
        Element element2 = new Element("element", ns);
        element2.setAttribute("name", "Participant_MsgFromReader");

        Element complexType2 = new Element("complexType", ns);

        element2.addContent(complexType2);
        Element sequence2 = new Element("sequence", ns);
        complexType2.addContent(sequence2);
        Element sender2 = new Element("element", ns);
        sender2.setAttribute("name", "sender");
        sender2.setAttribute("type", "tns:reader_name");
        sequence2.addContent(sender2);
        Element receiver2 = new Element("element", ns);
        receiver2.setAttribute("name", "receiver");
        receiver2.setAttribute("type", "tns:publisher_name");
        sequence2.addContent(receiver2);
        Element reply_to2 = new Element("element", ns);
        reply_to2.setAttribute("name", "reply_to");
        reply_to2.setAttribute("type", "tns:reader_name");
        sequence2.addContent(reply_to2);
        rootAddedTypesWSDL.addContent(element2);

    }

    /**
     * Aggiunge all'elelemnto root una linea del tipo <xs:element
     * name="reader_name" type="xs:string" fixed="JobSearcherName"/>
     * 
     * @param root
     * @param name
     * @param type
     * @param fixed
     */
    private void addElement(Element root, Namespace ns, String name,
            String type, String fixed) {
        Element nuovo = new Element("element", ns);
        nuovo.setAttribute("name", name);
        nuovo.setAttribute("type", type);
        nuovo.setAttribute("fixed", fixed);
        root.addContent(nuovo);

    }

    private void addPartnerLinkDeclaration(String pubRole, String readRole) {
        Element PartnerLinks;
        Element nuovo;
        String s;

        PartnerLinks = new Element("partnerLinks");
        rootBPEL.addContent(PartnerLinks);
        PartnerLinks.setNamespace(rootBPEL.getNamespace());

        //Publisher
        nuovo = new Element("partnerLink");
        nuovo.setNamespace(PartnerLinks.getNamespace());
        nuovo.setAttribute("name", "publisherPL");
        s = "lns:" + pubRole + readRole + "LinkLT";
        nuovo.setAttribute("partnerLinkType", s);
        nuovo.setAttribute("myRole", pubRole);
        nuovo.setAttribute("partnerRole", readRole);
        PartnerLinks.addContent(nuovo);
        //Reader
        nuovo = new Element("partnerLink");
        nuovo.setNamespace(PartnerLinks.getNamespace());
        nuovo.setAttribute("name", "readerPL");
        s = "lns:" + readRole + pubRole + "LinkLT";
        nuovo.setAttribute("partnerLinkType", s);
        nuovo.setAttribute("myRole", readRole);
        nuovo.setAttribute("partnerRole", pubRole);
        this.pubRole = pubRole;
        this.readRole = readRole;
        PartnerLinks.addContent(nuovo);

    }

    /**
     * Aggiunge al wsdl le seguenti righe <plnk:partnerLinkType
     * name="JobManagerJobSearcherLinkLT"> <plnk:role name="jobsManager">
     * <plnk:portType name="tns:publisherPT"/> </plnk:role> <plnk:role
     * name="jobsReader"> <plnk:portType name="tns:readerPT"/> </plnk:role>
     * </plnk:partnerLinkType>
     *  
     */

    private void addPartnerLinkTypeDeclaration() {
        Element PartnerLinkType = new Element("partnerLinkType");
        Namespace ns = Namespace.getNamespace("plnk",
                "http://schemas.xmlsoap.org/ws/2003/05/partner-link/");
        PartnerLinkType.setNamespace(ns);
        PartnerLinkType.setAttribute("name", pubRole + readRole + "LinkLT");

        //Publisher
        Element role = new Element("role", ns);
        role.setAttribute("name", pubRole);
        Element portType = new Element("portType", ns);
        portType.setAttribute("name", "tns:publisherPT");
        role.addContent(portType);
        PartnerLinkType.addContent(role);

        //      Reader
        Element role2 = new Element("role", ns);
        role2.setAttribute("name", readRole);
        Element portType2 = new Element("portType", ns);
        portType2.setAttribute("name", "tns:readerPT");
        role2.addContent(portType2);
        PartnerLinkType.addContent(role2);

        rootWSDL.addContent(PartnerLinkType);

    }

    /**
     * Prende in input una stringa del tipo jm/jobsManager:jobsManagerClass la
     * scompone e ne restituisce gli elementi se uno degli elementi manca ad
     * esempio per una stringa del tipo jm/: vengono restituiti gli elemengi di
     * defualt Se publisher  true si usano i valori di default per il publisher
     * altrimenti quelli per il reader in Result[0] viene restituito lo name
     * Result[1] viene restituito il role Result[2] viene restituito la classe
     * 
     * @param s
     * @return
     */

    private String[] ExtractNameRoleClass(String s, boolean publisher) {
        String[] result = new String[3];
        String name, role, classe;
        int start, end;
        end = s.indexOf("/");
        name = s.substring(0, end);
        String suff = "Publisher";
        if (publisher == false)
            suff = "Reader";

        if (name.length() == 0)
            name = "defaultAgentName" + suff;

        start = s.indexOf("/") + 1;
        end = s.indexOf(":");
        role = s.substring(start, end);
        if (role.length() == 0)
            name = "defaultAgentRole" + suff;

        start = s.indexOf(":") + 1;
        classe = s.substring(start);
        if (classe.length() == 0)
            name = "defaultClasseRole" + suff;

        result[0] = name;
        result[1] = role;
        result[2] = classe;

        return result;

    }

    /**
     * Restituisce un vettore contenente tutti i nodi con un certo nome tagName
     * individuati nella nodeList in input
     * 
     * @param nl
     *            Lista Nodi in cui ricercare
     * @param tagName
     *            Nome del tag da ricercare
     * @return Vettore con i nodi in nl aventi tagName
     */
    private Vector extractNodebyName(List l, String tagName) {
        Vector result = new Vector();
        org.jdom.Element tmp;
        for (int i = 0; i < l.size(); i++) {
            tmp = (org.jdom.Element) l.get(i);
            if (tmp.getName().equals(tagName))
                result.add(tmp);
        }
        return result;
    }

    /**
     * I param di un protocollo di comunicazione sono esperess nell'xml
     * intermedio secondo la notazione <mandatory-params>< <mandatory params>>
     * ontology = "job_search" content-language = "first order logic"
     * </mandatory-params>
     * 
     * @param s
     *            contenuto del tag mandatory param
     * @return stringhe relative all'ontologia ed al content language
     */

    private String[] extractValueFromMandatoryParam(String s) {
        String[] result = new String[2];
        String ontology;
        String cl;
        int start, end;
        start = s.indexOf("ontology") + 8;
        end = s.indexOf("CL");
        ontology = s.substring(start, end).trim();
        if (ontology.length() == 0)
            throw new IllegalArgumentException(
                    "Il file xml contiene errori relativi al tag mandatory-params (ontology)");
        ontology = ontology.replaceAll("=", "");
        ontology = ontology.replaceAll("\"", "");
        ontology = ontology.replaceAll("ontology", "");
        result[0] = ontology.trim();

        start = s.indexOf("CL");
        cl = s.substring((start+("CL".length()))).trim();
        if (cl.length() == 0)
            throw new IllegalArgumentException(
                    "Il file xml contiene errori relativi al tag mandatory-params (content-language)");
        cl = cl.replaceAll("=", "");
        cl = cl.replaceAll("\"", "");
        cl = cl.replaceAll("content-language", "");
        result[1] = cl.trim();
        return result;

    }

    /**
     * Estrae le informazioni relative agli agenti. Nella notazione xml
     * intermedia sono presenti i tag <agent-publisher>e <agent-reader>del
     * seguente tipo <agent-publisher
     * short-name="p">jm/jobsManager:jobsManagerClass </agent-publisher> La
     * funzione retituisce i dati in un formato proprio del linguaggio testuale
     * 
     * @param s
     *            Contenuto di <agent-publisher>o <agent-reader>
     * @return informazioni formattate
     */
    private String extractValueFromAgentTag(String s) {
        String result = "";
        String name, role, classe;
        int start, end;

        start = s.indexOf(">") + 1;
        end = s.indexOf("/");
        name = s.substring(start, end).trim();
        if (name.length() == 0)
            name = "df-agent-name";

        start = s.indexOf("/") + 1;
        end = s.indexOf(":");
        role = s.substring(start, end).trim();
        if (role.length() == 0)
            name = "df-role-name";

        start = s.indexOf(":") + 1;
        end = s.indexOf("<", start);
        classe = s.substring(start, end).trim();
        if (classe.length() == 0)
            name = "df-class-name";

        result = name + " " + role + " " + classe + " ";
        return result;

    }

}