package it.unige.disi.aumlbpel.transformers;

/**
 * @status da debuggare
 * 
 * @author Giovanni Casella it.unige.disi.aumlbpel.transformers
 *         IntermediateXML_TextualNotation.java Trasfoma un file xml in formato
 *         IntermediateXML nel documento wsdl adatto a supportare un processo
 *         bpel.
 */
import org.ietf.jgss.MessageProp;
import org.jdom.*;

import it.unige.disi.aumlbpel.bean.ProtocolMessage;

import javax.xml.parsers.DocumentBuilderFactory;
import org.jdom.output.*;

import java.io.*;
import java.util.List;
import java.util.Vector;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import java.io.IOException;

public class WSDL_BPEL_FOL {
    //root dei DOM rappresentanti i relativi Files
    private Element rootWSDL;
    private Element rootBPEL;
    private Element rootXML;
    
    //Document dei tre file
    private Document wsdlDocument;
    private Document bpelDocument;
    private Document xmlDocument;
    
    //Message Scambiati
    private Vector exchangedMessages = new Vector();
    
    private String ontology_name;
    private String content_language_name;
    private String protocol_name;
    private String publisher_name;
    private String publisher_role;
    private String reader_name;
    private String reader_role;
    
    
    private int formatting_tab_number = 0;
    
    public static void main(String[] argv) throws FileNotFoundException {
        if(argv.length!=3){
            System.out.println("Utilizzo: java -jar AUML2WSBPEL.jar nomefile.wsdl nomefile.bpel nomefile.fol \n");
        System.exit(1);
        }
        
        String nomeFileWsdlInput = argv[0];
        String NomeFileBpelInput = argv[1];
        String nomeFileFolOutput = argv[2];
        
        WSDL_BPEL_FOL trans = new WSDL_BPEL_FOL();
        trans.tranformWSDL_BPELtoFOL(nomeFileWsdlInput,
                NomeFileBpelInput, nomeFileFolOutput);
    }
    
    /**
     * Genera i documenti WSDL e Bpel a partire dall'Xml in input Il file XML
     * deve contenere codice XML Intermedio
     * 
     * @param nomeFileInput
     * @param nomeFileOutput
     * @throws FileNotFoundException
     */
    public void tranformWSDL_BPELtoFOL(String nomeWsdlInput,
            String nomeBpelInput, String nomeFolOut)
    throws FileNotFoundException {
        Vector elementVect;
        File wsdlFile = new File(nomeWsdlInput);
        File bpelFile = new File(nomeBpelInput);
              
        File FolOutFile = new File(nomeFolOut);
        File tmpFolFile = new File((FolOutFile.getParent()+File.separator+"tmp.txt"));
        System.out.println("TMP FILE:" + tmpFolFile.getPath());
        
        initializeDOM(wsdlFile, bpelFile, tmpFolFile);
        this.exchangedMessages = extractMessagesFromWsdl();
        
        FileOutputStream outS = new FileOutputStream(tmpFolFile);
        PrintStream ps = new PrintStream(outS);
        insertNotProcessElement(ps, rootXML);
        analyzeElement(ps, rootBPEL);
        //chiudo il process ed il main fragment aperto nella funzione insertNotProcessElement
        FormattingPrintln(ps, "))");
        try {
            cleanFol(tmpFolFile,FolOutFile);
            FolOutFile.deleteOnExit();
         } catch (FileNotFoundException e2) {
             // TODO Auto-generated catch block
             e2.printStackTrace();
         } 
        
    }
    
    /**
     * @param lastElement
     *  
     */
    private void insertNotProcessElement(PrintStream out, Element lastElement) {
        Element el;
        FormattingPrintln(out, "process(");
        formatting_tab_number++;
        FormattingPrintln(out, "parameters(");
        formatting_tab_number++;
        FormattingPrintln(out, "ontology('" + this.ontology_name + "'),");
        FormattingPrintln(out, "content_language('" + content_language_name
                + "')");
        formatting_tab_number--;
        FormattingPrintln(out, "),");
        FormattingPrintln(out, "%%% Definition of the protocol name %%%");
        FormattingPrintln(out, "protocol_name('" + protocol_name + "'),");
        FormattingPrintln(out, "%%% Definition of the agent publisher %%%");
        FormattingPrintln(out, "agent_publisher(short_name('" + publisher_role
                + "@address'),long_name('" + publisher_name + "/"
                + publisher_role + ":" + publisher_role+"Class" + "')),");
        FormattingPrintln(out, "%%% Definition of the agent reader %%%");
        FormattingPrintln(out, "agent_reader(short_name('" + reader_role
                + "@address'),long_name('" + reader_name + "/" + reader_role
                + ":" +reader_role+ "Class" + "')),");
        FormattingPrintln(out, "%%% Definition of the body contained in the WS_BPEL file %%%");
        FormattingPrintln(out, "main_fragment(");
                        
        
        
    }
    
   
    
    private void analyzeElement(PrintStream out, Element elementToAnalyze) {
        Vector nodeVect;
        List children;
        children = elementToAnalyze.getChildren();
        
        Element t;
        for (int i = 0; i < children.size(); i++) {
            t = (Element) children.get(i);
            
            if (t.getName().equals("receive")) {
                processReceive(out, t);
                
            } else if (t.getName().equals("sequence")) {
                processSequenceAct(out, t);
                
            } else if (t.getName().equals("switch")) {
                processSwitchAct(out, t);
            } else if (t.getName().equals("case")) {
                processCaseAct(out, t);
            } else if (t.getName().equals("while")) {
                processLoopAct(out, t);
            } else if (t.getName().equals("if")) {
                processIfAct(out, t);
            }
            else if (t.getName().equals("then")) {
                analyzeElement(out, t);
            }
            
            
        }
        return;
    }
    
    /**
     * @param t
     */
    private void processIfAct(PrintStream out, Element t) {
        Element tmp;
        formatting_tab_number++;
        FormattingPrintln(out,"if_then(no_guard,");
        
        analyzeElement(out, t);
        FormattingPrintln(out,"),");
        formatting_tab_number--;
    }
    
    private void processLoopAct(PrintStream out, Element t) {
        Element tmp;
        formatting_tab_number++;
        FormattingPrintln(out, "while(no_guard,");
        analyzeElement(out, t);
        FormattingPrintln(out, "),");
        formatting_tab_number--;
    }
    
    /**
     * @param t
     */
    private void processCaseAct(PrintStream out, Element t) {
        formatting_tab_number++;
        FormattingPrintln(out, "case(no_guard,");
        analyzeElement(out, t);
        
        FormattingPrintln(out, "),");
        formatting_tab_number--;
    }
    
    /**
     * @param t
     */
    private void processSwitchAct(PrintStream out, Element t) {
        
        formatting_tab_number++;
        FormattingPrintln(out, "switch([");
        
        analyzeElement(out, t);
        FormattingPrintln(out, "]),");  
        formatting_tab_number--;
        
    }
    
    /**
     * @param t
     */
    private void processReceive(PrintStream out, Element el) {
        
        String s;
        s = el.getAttributeValue("operation");
        //Estraggo il numnero del messaggio dalla stringa
        //RCV_Mess_1
        s = s.substring(9, s.length());
        int msgNumber = Integer.parseInt(s);
        ProtocolMessage pm = (ProtocolMessage) exchangedMessages.get((msgNumber-1));
        String readerOp;
        String senderStr;
        if (pm.getSender().equals("r"))
            senderStr = "reader";
        else
            senderStr = "publisher";
        
        if (senderStr.equals("publisher"))
            readerOp = "receive";
        else
            readerOp = "send";
        //send(msg('REQUEST', 'accepted_payment_methods')),
        s = readerOp + "(msg('" + pm.getPerformative().toUpperCase() + "','"
        + pm.getContent() + "')),";
        FormattingPrintln(out, s);
        
    }
    
    /**
     * @param t
     * @param lastInserted
     */
    private void processSequenceAct(PrintStream out, Element elementToAnalyze) {
        Element tmp;
        formatting_tab_number++;
        FormattingPrintln(out, "seq([");
        
        analyzeElement(out, elementToAnalyze);
        FormattingPrintln(out, "]),");
        formatting_tab_number--;
    }
    
    /**
     * @param wsdlFile
     */
    private Vector extractMessagesFromWsdl() {
        Vector messages = new Vector();
        
        Vector v = extractNodebyName(rootWSDL.getChildren(), "types");
        
        Element TypesRoot = (Element) v.get(0);
        
        extractProtocolInfo(TypesRoot);
        
        Vector messagesTags = extractNodebyName(TypesRoot.getChildren(),
        "message");
        Element tmp;
        Element part;
        Vector v2;
        Element tmp2;
        Vector messageParts;
        for (int i = 0; i < messagesTags.size(); i++) {
            ProtocolMessage pm = new ProtocolMessage();
            tmp = (Element) messagesTags.get(i);
            pm.setName(tmp.getAttributeValue("name"));
            
            messageParts = extractNodebyName(tmp.getChildren(), "part");
            for (int j = 0; j < messageParts.size(); j++) {
                part = (Element) messageParts.get(j);
                if (part.getAttributeValue("name").equals("performative")) {
                    String perf = part.getAttributeValue("element");
                    //Tagli il _ca da request_ca
                    perf = perf.substring(0, (perf.length() - 3));
                    pm.setPerformative(perf);
                }
                if (part.getAttributeValue("name").equals("Content")) {
                    //System.out.println("Content to find: " +
                    // part.getAttributeValue("element"));
                    //System.out.println("Content found: " +
                    // findMessageContent(TypesRoot,
                    // part.getAttributeValue("element")));
                    pm.setContent(findMessageContent(TypesRoot, part
                            .getAttributeValue("element")));
                    
                }
                if (part.getAttributeValue("name").equals("Participant")) {
                    if (part.getAttributeValue("type").equals(
                    "tns:Participant_MsgFromPublisher")) {
                        pm.setSender("p");
                    } else
                        pm.setSender("r");
                    
                }
                
            }
            messages.add(i,pm);
            pm.getContent();
           /*
            System.out.println("*** " + pm.getName());
            System.out.println("*** " + pm.getPerformative());
            System.out.println(">>> " + pm.getContent());
            System.out.println(" < < < " + pm.getSender());
            System.out.println(" < < < " + pm.getContent());
           */
            
        }
        
        return messages;
    }
    
    /**
     * @param typesRoot
     * @param attributeValue
     * @return
     */
    private String findMessageContent(Element typesRoot, String contentType) {
        Vector vElements = extractNodebyName(typesRoot.getChildren(), "element");
        Element tmp;
        contentType = contentType.substring(4, contentType.length());
        for (int i = 0; i < vElements.size(); i++) {
            tmp = (Element) vElements.get(i);
            if (tmp.getAttributeValue("name").equals(contentType)) {
                return tmp.getAttributeValue("fixed");
                
            }
        }
        // TODO Auto-generated method stub
        return null;
    }
    
    /**
     * @param typesRoot
     */
    private void extractProtocolInfo(Element typesRoot) {
        List l = typesRoot.getChildren();
        org.jdom.Element tmp;
        for (int i = 0; i < l.size(); i++) {
            tmp = (org.jdom.Element) l.get(i);
            if ((tmp.getAttributeValue("name") != null)
                    && (tmp.getAttributeValue("name").equals("ontology_name")))
                this.ontology_name = tmp.getAttributeValue("fixed");
            if ((tmp.getAttributeValue("name") != null)
                    && (tmp.getAttributeValue("name")
                            .equals("content_language_name")))
                this.content_language_name = tmp.getAttributeValue("fixed");
            if ((tmp.getAttributeValue("name") != null)
                    && (tmp.getAttributeValue("name").equals("protocol_name")))
                this.protocol_name = tmp.getAttributeValue("fixed");
            
            if ((tmp.getAttributeValue("name") != null)
                    && (tmp.getAttributeValue("name").equals("publisher_name")))
                this.publisher_name = tmp.getAttributeValue("fixed");
            if ((tmp.getAttributeValue("name") != null)
                    && (tmp.getAttributeValue("name").equals("publisher_role")))
                this.publisher_role = tmp.getAttributeValue("fixed");
            if ((tmp.getAttributeValue("name") != null)
                    && (tmp.getAttributeValue("name").equals("reader_name")))
                this.reader_name = tmp.getAttributeValue("fixed");
            if ((tmp.getAttributeValue("name") != null)
                    && (tmp.getAttributeValue("name").equals("reader_role")))
                this.reader_role = tmp.getAttributeValue("fixed");
        }
        
    }
    
    private Vector extractNodebyName(NodeList nl, String tagName) {
        Vector result = new Vector();
        Node tmp;
        for (int i = 0; i < nl.getLength(); i++) {
            tmp = nl.item(i);
            if (tmp.getNodeName().equals(tagName))
                result.add(tmp);
        }
        return result;
    }
    
    private Vector extractNodebyName(List l, String tagName) {
        Vector result = new Vector();
        org.jdom.Element tmp;
        for (int i = 0; i < l.size(); i++) {
            tmp = (org.jdom.Element) l.get(i);
            if (tmp.getName().equals(tagName))
                result.add(tmp);
            
        }
        return result;
    }
    
    /**
     * Scrive sui rispettivi File l'albero DOM rapprensetante il documento Bpel
     * e quello che rapprensenta il documento Wsdl
     * 
     * @param nomeFileWsdlOutput
     * @param nomeFileBpelOutput
     */
    private void writeDoms(String nomeFileWsdlOutput, String nomeFileBpelOutput) {
        File outputWsdlFile = new File(nomeFileWsdlOutput);
        File outputBpelFile = new File(nomeFileBpelOutput);
        OutputStream outWsdl, outBpel;
        try {
            outWsdl = new FileOutputStream(outputWsdlFile);
            outBpel = new FileOutputStream(outputBpelFile);
            //          SCRITTURA OUTPUT Wsdl e Bpel
            XMLOutputter outputter = new XMLOutputter();
            outputter.output(rootWSDL, outWsdl);
            outputter.output(rootBPEL, outBpel);
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }
        
    }
    
    private void initializeDOM(File wsdlFile, File bpelFile, File XmlOutFile) {
        //Inizilaizzazione gestori DOM
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        Vector nodeVect;
        SAXBuilder builder = new SAXBuilder();
        try {
            Document outputDom = new Document();
            // APERUTURA FILE DOM WSDL E BPEL
            this.wsdlDocument = builder.build(wsdlFile);
            rootWSDL = wsdlDocument.getRootElement();
            this.bpelDocument = builder.build(bpelFile);
            rootBPEL = bpelDocument.getRootElement();
        } catch (JDOMException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }
    
    /**
     * Stampa la stringa s formattandola correttamente e andando a capo
     * 
     * @param out
     * @param s
     */
    private void FormattingPrintln(PrintStream out, String s) {
        for (int i = 0; i < formatting_tab_number; i++) {
            s = "   " + s;
            //s="\t"+s;}
            
        }
        out.println(s);
        
    }
    
    /**
     * @param outputFolFile
     * Ripulisco il file eliminando le , vicino alle parentesi ,)
     * @throws FileNotFoundException
     */
    private void cleanFol(File outputFolFile, File cleanoutputFolFile) throws FileNotFoundException {
        
	    FileOutputStream tmpos1 = new FileOutputStream(cleanoutputFolFile);
        PrintStream tmp_wr = new PrintStream(tmpos1);
       System.out.println("WRITING ON "+cleanoutputFolFile.getPath());
        
        
        
        FileInputStream fis1 = new FileInputStream(outputFolFile);
		InputStreamReader isr1=new InputStreamReader (fis1);
		BufferedReader rdr1 = new BufferedReader(isr1);
		
		
		String lastLine=" ";
		String thisLine="";
		int removeInd;
		try {
		    thisLine=rdr1.readLine();
		    while(thisLine!=null){
		        //Ripulisco la linea da eventuali elementi ,) oppure ,]
		        //String r = Pattern.compile(",\\)").matcher(l).replaceFirst(")");
		        //r=Pattern.compile(",]").matcher(l).replaceFirst("]");
		        if((thisLine.trim().charAt(0)==')')||(thisLine.trim().charAt(0)==']')) {
		          // if(lastLine.charAt(lastLine.length()-1)==',')
		          if(lastLine.endsWith(","))
			        {
		              char[] linA=lastLine.toCharArray();
		              linA[linA.length-1]=' ';
		              lastLine="";
		              for(int i=0;i<linA.length;i++)
		              lastLine=lastLine+linA[i];
		            
		              
			        }
		       	}      
		        
		        if(!lastLine.equals(" ")) tmp_wr.println(lastLine);
		        lastLine=thisLine;
		        thisLine=rdr1.readLine();
		    
		    }
		    if(!lastLine.equals(" ")) tmp_wr.println(lastLine);
	       
		    
		    
		    
	        
        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

        
    }
    
}